# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2012-2018, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect, sys
import os.path as osp

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.lib.disassemble import dis
import trepan.lib.file as Mfile
from xdis.load import load_module
from trepan.processor.cmd_addrlist import parse_addr_list_cmd

# From importlib.util
DEBUG_BYTECODE_SUFFIXES = [".pyc"]
OPTIMIZED_BYTECODE_SUFFIXES = [".pyo"]
_PYCACHE = "__pycache__"


def cache_from_source(path, debug_override=None):
    """Given the path to a .py file, return the path to its .pyc/.pyo file.

    The .py file does not need to exist; this simply returns the path to the
    .pyc/.pyo file calculated as if the .py file were imported.  The extension
    will be .pyc unless sys.flags.optimize is non-zero, then it will be .pyo.

    If debug_override is not None, then it must be a boolean and is used in
    place of sys.flags.optimize.

    If sys.implementation.cache_tag is None then NotImplementedError is raised.

    """
    debug = not sys.flags.optimize if debug_override is None else debug_override
    if debug:
        suffixes = DEBUG_BYTECODE_SUFFIXES
    else:
        suffixes = OPTIMIZED_BYTECODE_SUFFIXES
        pass
    head, tail = osp.split(path)
    base_filename, sep, _ = tail.partition(".")
    if not hasattr(sys, "implementation"):
        # Python <= 3.2
        raise NotImplementedError("No sys.implementation")
    tag = sys.implementation.cache_tag
    if tag is None:
        raise NotImplementedError("sys.implementation.cache_tag is None")
    filename = "".join([base_filename, sep, tag, suffixes[0]])
    if head.endswith(_PYCACHE):
        return osp.join(head, filename)
    return osp.join(head, _PYCACHE, filename)


class DisassembleCommand(DebuggerCommand):
    """**disassemble** [*thing*]

**disassemble** [*address-range*]

Disassembles bytecode. See `help syntax range` for what can go in a list range.

Without arguments, print lines starting from where the last list left off
since the last entry to the debugger. We start off at the location indicated
by the current stack.

in addition you can also use:

  - a '.' for the location of the current frame

  - a '-' for the lines before the last list

  - a '+' for the lines after the last list

With a class, method, function, pyc-file, code or string argument
disassemble that.

Examples:
--------
::

   disassemble    # Possibly current frame
   disassemble +                    # Same as above
   disassemble os.path              # Disassemble all of os.path # xxx
   disassemble os.path.normcase()   # Disaasemble just method os.path.normcase
   disassemble 3                    # Disassemble starting from line 3
   disassemble 3, 10                # Disassemble lines 3 to 10
   disassemble *0, *10              # Disassemble offset 0..10 (10 not included)
   disassemble myprog.pyc           # Disassemble file myprog.pyc

See also:
---------

`help syntax arange`, `deparse`, `list`, `info pc`.
"""

    aliases = ("disasm",)  # Note: we will have disable
    short_help = "Disassemble Python bytecode"

    DebuggerCommand.setup(
        locals(), category="data", max_args=2
    )

    def run(self, args):
        proc = self.proc

        # FIXME: add a setting for assembler list size
        listsize = 4

        opts = {
            "highlight": self.settings["highlight"],
            "start_line": 1,
            "end_line": None,
            "start_offset": None,
            "end_offset": None,
            "relative_pos": False,
            "asm_format" : self.settings["asmfmt"]
        }

        curframe = proc.curframe
        if curframe:
            line_no = inspect.getlineno(curframe)
            opts["start_line"] = line_no - 1
            opts["end_line"] = line_no + 1

        do_parse = True
        if len(args) == 2:
            if args[1].endswith("()"):
                eval_args = args[1][:-2]
            else:
                eval_args = args[1]
            try:
                obj = self.proc.eval(eval_args, show_error=False)
            except:
                obj = None
            else:
                if (
                    inspect.ismethod(obj)
                    or inspect.isfunction(obj)
                    or inspect.isgeneratorfunction(obj)
                    or inspect.isgenerator(obj)
                    or inspect.isframe(obj)
                    or inspect.iscode(obj)
                ):
                    opts["start_offset"] = 0
                    last_is_offset = is_offset = True
                    start = 0
                    opts["start_line"] = 0
                    opts["end_line"] = last = None
                    do_parse = False
                    bytecode_file = None

        if do_parse:
            (
                bytecode_file,
                start,
                is_offset,
                last,
                last_is_offset,
                obj,
            ) = parse_addr_list_cmd(proc, args, listsize)
            if bytecode_file is None:
                return

        if is_offset:
            opts["start_offset"] = start
        else:
            opts["start_line"] = start
        if last_is_offset:
            opts["end_offset"] = last
        else:
            opts["end_line"] = last

        if not obj and (
            bytecode_file
            and (not bytecode_file.endswith(".pyo") or bytecode_file.endswith("pyc"))
        ):
            # bytecode_file may be a source file. Try to tun it into a bytecode file for diassembly.
            bytecode_file = cache_from_source(bytecode_file)
            if bytecode_file and Mfile.readable(bytecode_file):
                self.msg("Reading %s ..." % bytecode_file)
                (
                    version,
                    timestamp,
                    magic_int,
                    obj,
                    is_pypy,
                    source_size,
                    sip_hash,
                ) = load_module(bytecode_file)
            elif not curframe:
                self.errmsg("No frame selected.")
                return
            else:
                try:
                    obj = self.proc.eval(args[1])
                    opts["start_line"] = -1
                except:
                    self.errmsg(
                        ("Object '%s' is not something we can" + " disassemble.")
                        % args[1]
                    )
                    return

        # We now have all  information. Do the listing.
        (obj, proc.list_offset) = dis(
            self.msg, self.msg_nocr, self.section, self.errmsg, obj, **opts
        )
        return False


# Demo it
if __name__ == "__main__":

    # FIXME: make sure the below is in a unit test
    def doit(cmd, args):
        proc = cmd.proc
        proc.current_command = " ".join(args)
        cmd.run(args)

    from trepan.processor.command import mock

    d, cp = mock.dbg_setup()

    cp.list_object = cp.curframe = inspect.currentframe()
    command = DisassembleCommand(cp)
    prefix = "-" * 20 + " disassemble "

    print(prefix + "os.path")
    doit(command, ["dissassemble", "cp.errmsg()"])

    print(prefix + "cp.errmsg()")
    doit(command, ["dissassemble", "cp.errmsg()"])

    print(prefix + "cp.errmsg()")
    doit(command, ["dissassemble", "cp.errmsg()"])

    # print(prefix)
    # doit(command, ['disassemble']) # no good

    # print(prefix + 'me')
    # doit(command, ['disassemble', 'me()']) # reports invalid function correctly

    print(prefix + "*0 +248")
    doit(command, ["disassemble", "*0,", "+248"])

    # print(prefix + '+ 2-1')
    # doit(command, ['disassemble', '+', '2-1']) # not valid?

    # print(prefix + '- 1')
    # doit(command, ['disassemble', '-', '1']) # not working

    print(prefix + ".")
    doit(command, ["disassemble", "."])

    __file__ = "./disassemble.py:21"
    doit(command, ["disassemble", "%s:21" % __file__])

    # bytecode_file = cache_from_source(__file__)
    # print(bytecode_file)
    # if bytecode_file:
    #     doit(command, ["disassemble", bytecode_file + ":22,28"])

    doit(command, ["disassemble", "*15,", "*25"])
    doit(command, ["disassemble", "30"])
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009-2010, 2013, 2015, 2017-2018, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.lib.printing import printf
from trepan.processor.complete import complete_identifier


class PCommand(DebuggerCommand):
    """**print** *expression*

Print the value of the expression. Variables accessible are those of the
environment of the selected stack frame, plus globals.

The expression may be preceded with */fmt* where *fmt* is one of the
format letters 'c', 'x', 'o', 'f', or 's' for chr, hex, oct,
float or str respectively.

If the length output string large, the first part of the value is
shown and `...` indicates it has been truncated.

See also:
---------

 `pp` and `examine` for commands which do more in the way of formatting.
"""

    aliases = ("print", "pr")
    short_help = "Print value of expression EXP"

    complete = complete_identifier

    DebuggerCommand.setup(locals(), category="data", need_stack=True, min_args=1)

    def run(self, args):
        if len(args) > 2 and "/" == args[1][0]:
            fmt = args[1]
            del args[1]
        else:
            fmt = None
            pass
        arg = " ".join(args[1:])
        try:
            val = self.proc.eval(arg)
            if fmt:
                val = printf(val, fmt)
                pass
            self.msg(self.proc._saferepr(val))
        except:
            pass


if __name__ == "__main__":
    import inspect
    from trepan.processor.command import mock

    d, cp = mock.dbg_setup()
    cp.curframe = inspect.currentframe()
    command = PCommand(cp)
    me = 10

    command.run([command.name, "me"])
    command.run([command.name, "/x", "me"])
    command.run([command.name, "/o", "me"])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2015 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd
from trepan.lib import complete as Mcomplete
from trepan.processor import frame as Mframe


class InfoCode(Mbase_subcmd.DebuggerSubcommand):
    """**info code** [ *frame-number* | *pyton code object* ]

Show the detailed information for the Python code object in
*frame-number* or the current frame if *frame-number* is not specified.

Specific information includes:

* the number of arguments (not including * or ** args)

* constants used in the bytecode

* name of file in which this code object was created

* number of first line in Python source code

* name with which this code object was defined

See also:
---------

`info frame`, `info locals`, `info file`
"""

    min_abbrev = 2
    max_args = 2
    need_stack = True
    short_help = '''Show detailed info about the Python code object'''

    def complete(self, prefix):
        proc_obj = self.proc
        low, high = Mframe.frame_low_high(proc_obj, None)
        ary = [str(low+i) for i in range(high-low+1)]
        # FIXME: add in Thread names
        return Mcomplete.complete_token(ary, prefix)

    def run(self, args):
        proc = self.proc
        frame = proc.curframe
        if not frame:
            self.errmsg("No frame selected.")
            return False

        frame_num = None
        frame_num = None
        if len(args) == 1:
            try:
                frame_num = int(args[0])
                i_stack = len(proc.stack)
                if frame_num < -i_stack or frame_num > i_stack-1:
                    self.errmsg(('a frame number number has to be in the range %d to %d.' +
                                 ' Got: %d (%s).') % (-i_stack, i_stack-1,
                                                      frame_num, args[0]))
                    return False
                frame = proc.stack[frame_num][0]
                code = frame.f_code
            except:
                try:
                    code = eval(args[0], frame.f_globals, frame.f_locals)
                except:
                    self.errmsg('%s is not a evaluable as a code object or frame number.' %
                                 args[0])
                    return False
                # if not inspect.iscode(code):
                #     self.errmsg('%s is not a code object' % code)
                pass

        else:
            frame_num = proc.curindex
            code = frame.f_code

        mess = 'Code for Frame %d' % frame_num if frame_num is not None else 'Code Info'
        self.section(mess)
        self.msg("  name: %s" % code.co_name)
        self.msg("  number of arguments: %d" % code.co_argcount)
        self.msg("  number of locals: %d" % code.co_nlocals)
        self.msg("  maximum stack size %s" % code.co_stacksize)
        self.msg("  first line number: %s" % code.co_firstlineno)
        self.msg("  is%s optimized" % ("" if (code.co_flags & 1) == 1 else " not"))
        self.msg("  new local namespace? %s" % ("yes" if (code.co_flags & 2) == 1 else " no"))
        self.msg("  has%s *args" % ("" if (code.co_flags & 4) == 1 else " no"))
        self.msg("  has%s **args" % ("" if (code.co_flags & 8) == 1 else " no"))
        maxwidth = self.settings['width'] // 2
        for name, field in [('Constants', 'co_consts'),
                            ('Variable names', 'co_varnames'),
                            ('Local Variables', 'co_names')
                            ]:
            vals = [proc._saferepr(x, maxwidth) for x in getattr(code, field)]
            if vals:
                self.section(name)
                m = self.columnize_commands(vals)
                self.msg_nocr(m)

        return False
    pass

if __name__ == '__main__':
    from trepan.processor.command import mock, info as Minfo
    d, cp = mock.dbg_setup()
    i = Minfo.InfoCommand(cp)
    sub = InfoCode(i)
    import inspect
    cp.curframe = inspect.currentframe()
    sub.run([])



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013-2015 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger Socket Input/Output Interface."""

import socket

from trepan.lib import default as Mdefault
from trepan.inout import tcpfns as Mtcpfns
from trepan.inout.base import DebuggerInOutBase
from trepan.misc import option_set


class TCPClient(DebuggerInOutBase):
    """Debugger Client Input/Output Socket."""

    DEFAULT_INIT_OPTS = {'open': True}

    def __init__(self, inout=None, opts=None):
        get_option = lambda key: option_set(opts, key,
                                            Mdefault.CLIENT_SOCKET_OPTS)
        self.inout = None
        self.addr = None
        self.buf = ''
        self.line_edit = False  # Our name for GNU readline capability
        self.state = 'disconnected'
        if inout:
            self.inout = inout
        elif get_option('open'):
            self.open(opts)
            pass
        return

    def close(self):
        """ Closes both input and output """
        if self.inout:
            self.inout.close()
            pass
        self.state = 'disconnnected'
        return

    def open(self, opts=None):

        get_option = lambda key: option_set(opts, key,
                                            Mdefault.CLIENT_SOCKET_OPTS)

        HOST = get_option('HOST')
        PORT = get_option('PORT')
        self.inout = None
        for res in socket.getaddrinfo(HOST, PORT, socket.AF_UNSPEC,
                                      socket.SOCK_STREAM):
            af, socktype, proto, canonname, sa = res
            try:
                self.inout = socket.socket(af, socktype, proto)
                self.state = 'connected'
            except socket.error:
                self.inout = None
                self.state = 'disconnected'
                continue
            try:
                self.inout.connect(sa)
            except socket.error:
                self.inout.close()
                self.inout = None
                continue
                break
        if self.inout is None:
            raise IOError('could not open client socket on port %s' %
                          PORT)
        return

    def read_msg(self):
        """Read one message unit. It's possible however that
        more than one message will be set in a receive, so we will
        have to buffer that for the next read.
        EOFError will be raised on EOF.
        """
        if self.state == 'connected':
            if 0 == len(self.buf):
                self.buf = self.inout.recv(Mtcpfns.TCP_MAX_PACKET)
                if 0 == (self.buf):
                    self.state = 'disconnected'
                    raise EOFError
                pass
            self.buf, data = Mtcpfns.unpack_msg(self.buf)
            return data.decode('utf-8')
        else:
            raise IOError("read_msg called in state: %s." % self.state)

    def write(self, msg):
        """ This method the debugger uses to write a message unit."""
        # FIXME: do we have to check the size of msg and split output?
        return self.inout.send(Mtcpfns.pack_msg(msg))

    pass

# Demo
if __name__=='__main__':
    inout = TCPClient(opts={'open': False})
    import sys
    if len(sys.argv) > 1:
        print('Connecting...', end=' ')
        inout.open()
        print('connected.')
        while True:
            line = input('nu? ')
            if len(line) == 0: break
            try:
                line = inout.writeline(line)
                print("Got: ", inout.read_msg().rstrip('\n'))
            except EOFError:
                break
            pass
        pass
    inout.close()
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2012-2015 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Things related to eval/exec"""

# extract the "expression" part of a line of source code.
#
import re


def extract_expression(text):
    if re.search("^\s*(?:if|elif)\s+", text):
        text = re.sub("^\s*(?:if|elif)\s+", "", text)
        text = re.sub(":(?:\s+.*$|$)", "", text)
    elif re.search("^\s*assert\s+.*", text):
        # EXPR in : assert EXPRESSION:
        text = re.sub("^\s*assert\s+", "", text)
    elif re.search("^\s*(?:while)\s+", text):
        text = re.sub("^\s*(?:while)\s+", "", text)
        text = re.sub(":(?:\s+.*$|$)", "", text)
    elif re.search("^\s*return\s+", text):
        # EXPRESION in: return EXPRESSION
        text = re.sub("^\s*return\s+", "", text)
    elif re.search("^\s*for\s+.+\s+in\s+.*:", text):
        # EXPRESION in: for VAR in EXPRESSION:
        text = re.sub("^\s*for\s+.+\s+in\s+", "", text)
        text = re.sub(":.*$", "", text)
    elif re.search("^\s*and\s+.*", text):
        # EXPRESION in: and EXPRESSION
        text = re.sub("^\s*and\s+", "", text)
    elif re.search("^\s*or\s+.*", text):
        # EXPRESION in: and EXPRESSION
        text = re.sub("^\s*or\s+", "", text)
    elif re.search("\s*[A-Za-z_][A-Za-z0-9_\[\]]*\s*=[^=>]", text):
        # RHS of an assignment statement.
        text = re.sub("^\s*[A-Za-z_][A-Za-z0-9_\[\]]*\s*=\s*", "", text)
        pass
    return text


# Demo it
if __name__ == "__main__":
    for stmt in (
        "if condition(x):",
        "elif _is_magic(name):",
        "while expression:",
        "for i in range(3):",
        "and x > 3",
        "or y < 3",
        "abc = 123",
        "return return_value",
        "nothing_to_be.done",
    ):
        print(extract_expression(stmt))
        pass
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand
from pyficache import cache_code_lines


class InfoOffsets(DebuggerSubcommand):
    """**info offsets** [*function-file-or-module*]

Show line offset information for a function, module, or a file.

If no location is given, use the the current frame.

Examples
--------

    (trepan3k) info offsets
    Offset - line number table for offsets.py
           0:  16    40:  23    72:  94   108:  98   136: 101   164: 104   188: 107
          16:  19    48:  45    84:  95   114:  99   144: 102   174: 105
          28:  20    64:  93    96:  96   128: 100   154: 103   184: 106

See also:
---------
`info line`, `info program`, `info frame`"""

    min_abbrev = 2
    max_args = 2
    need_stack = False
    short_help = "Show line offset information for a file or module"

    def run(self, args):
        """Current line number in source file"""
        # info line <loc>
        if len(args) == 0:
            curframe = self.proc.curframe
            if not curframe:
                self.errmsg("No line number information available.")
                return

            # No line number. Use current frame line number
            filename = curframe.f_code.co_filename
            file_info = cache_code_lines(filename, toplevel_only=False, include_offsets=True)
            if file_info:
                self.section("Offset - line number table for %s" % filename)
                offsets = ["@%4d:%4d" % (offset, line) for offset, line in file_info.linestarts.items()]
                m = self.columnize_commands(list(sorted(offsets)))
                self.msg(m)
            else:
                self.errmsg(
                    "haven't recorded info for offset file %s" % filename
                )
                pass
            pass
        else:
            self.errmsg("Passing a filename, function, or module not completed yet...")
        return
    pass


if __name__ == "__main__":
    from trepan.processor.command import mock
    from trepan.processor.command.info import InfoCommand
    from trepan.debugger import Trepan

    d = Trepan()
    d, cp = mock.dbg_setup(d)
    i = InfoCommand(cp)
    sub = InfoOffsets(i)
    import inspect
    cp.curframe = inspect.currentframe()
    for width in (80, 200):
        sub.settings["width"] = width
        sub.run([])
        pass
    pass



#!/usr/bin/env python3
'Unit test for trepan.processor.command.cmdfns'
import unittest

from trepan.processor import cmdfns as Mcmdfns


class TestCommandHelper(unittest.TestCase):

    def setUp(self):
        self.errors = []
        return

    def errmsg(self, msg):
        self.errors.append(msg)
        return

    def test_get_an_int(self):
        self.assertEqual(0, Mcmdfns.get_an_int(self.errmsg, '0', 'foo', 0))
        self.assertEqual(0, len(self.errors))
        self.assertEqual(6, Mcmdfns.get_an_int(self.errmsg, '6*1', 'foo', 5))
        self.assertEqual(0, len(self.errors))
        self.assertEqual(None, Mcmdfns.get_an_int(self.errmsg, '0',
                                                  '0 is too small', 5))
        self.assertEqual(1, len(self.errors))
        self.assertEqual(None, Mcmdfns.get_an_int(self.errmsg, '4+a',
                                                  '4+a is invalid', 5))
        self.assertEqual('4+a is invalid', self.errors[-1])
        return

    def test_get_int(self):
        self.assertEqual(1, Mcmdfns.get_int(self.errmsg, '1', 5))
        self.assertEqual(3, Mcmdfns.get_int(self.errmsg, '1+2', 5))
        self.assertEqual(5, Mcmdfns.get_int(self.errmsg, None, 5))
        self.assertEqual(1, Mcmdfns.get_int(self.errmsg, None))
        self.assertRaises(ValueError, Mcmdfns.get_int,
                          *(self.errmsg, 'Foo', 5))
        return

    def test_get_onoff(self):
        for arg in ('1', 'on'):
            self.assertEqual(True, Mcmdfns.get_onoff(self.errmsg, arg))
            pass
        for arg in ('0', 'off'):
            self.assertEqual(False, Mcmdfns.get_onoff(self.errmsg, arg))
            pass
        for result in (True, False):
            self.assertEqual(result, Mcmdfns.get_onoff(self.errmsg, None,
                                                       result))
            pass
        self.assertRaises(ValueError, Mcmdfns.get_onoff, *(self.errmsg,
                                                           'Foo'))
        return

    def test_want_different_line(self):
        for cmd, default, expected in [
        ('s+', False, True),
        ('s-', True,  False),
        ('s',  False, False),
        ('n',  True,  True) ]:
            self.assertEqual(expected,
                             Mcmdfns.want_different_line(cmd, default),
                             cmd)
            pass
        return
    pass

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#   Copyright (C) 2015, 2017, 2020 Rocky Bernstein <rocky@gnu.org>
"""Breakpoints as used in a debugger.

This code is a rewrite of the stock python bdb.Breakpoint"""

__all__ = ["BreakpointManager", "Breakpoint"]

import os.path


class BreakpointManager(object):
    """Manages the list of Breakpoints.

    Breakpoints are indexed by number in the `bpbynumber' list, and
    through a (file,line) tuple which is a key in the `bplist'
    dictionary. If the breakpoint is a function it is in `fnlist' as
    well.  Note there may be more than one breakpoint per line which
    may have different conditions associated with them.
    """
    def __init__(self):
        self.reset()
        return

    def bpnumbers(self):
        '''Returns a list of strings of breakpoint numbers'''
        return ["%d" % bp.number for bp in self.bpbynumber if
                bp is not None]

    def get_breakpoint(self, i):
        if isinstance(i, str):
            try:
                i = int(i)
            except ValueError:
                return(False, 'Breakpoint value %r is not a number.' % i,
                       None)
            pass
        if 1 == len(self.bpbynumber):
            return (False, 'No breakpoints set.', None)
        elif i >= len(self.bpbynumber) or i <= 0:
            return (False, 'Breakpoint number %d out of range 1..%d.' %
                    (i, len(self.bpbynumber)-1), None)
        bp = self.bpbynumber[i]
        if bp is None:
            return (False, 'Breakpoint %d previously deleted.' % i, None)
        return (True, None, bp)

    def add_breakpoint(self, filename, lineno, offset, temporary=False, condition=None,
                       func=None):

        bpnum = len(self.bpbynumber)
        if filename: filename  = os.path.realpath(filename)
        brkpt = Breakpoint(bpnum, filename, lineno, temporary, condition,
                           func, offset)
        # Build the internal lists of breakpoints
        self.bpbynumber.append(brkpt)
        if (filename, lineno) in self.bplist:
            self.bplist[filename, lineno].append(brkpt)
        else:
            self.bplist[filename, lineno] = [brkpt]
            pass
        if func and offset in [None, -1]:
            if func in self.fnlist:
                self.fnlist[func].append(brkpt)
            else:
                self.fnlist[func] = [brkpt]
                pass
        return brkpt

    def delete_all_breakpoints(self):
        bp_list = []
        for bp in self.bpbynumber:
            if bp:
                bp_list.append(str(bp.number))
                self.delete_breakpoint(bp)
                pass
        if not bp_list:
            return 'There are no breakpoints'
        else:
            return 'Deleted breakpoints %s' % ', '.join(bp_list)
        return

    def delete_breakpoint(self, bp):
        " remove breakpoint `bp'"
        bpnum = bp.number
        self.bpbynumber[bpnum] = None   # No longer in list
        index = (bp.filename, bp.line)
        if index not in self.bplist: return False
        self.bplist[index].remove(bp)
        if not self.bplist[index]:
            # No more breakpoints for this file:line combo
            del self.bplist[index]
        return True

    def delete_breakpoint_by_number(self, bpnum):
        "Remove a breakpoint given its breakpoint number."
        success, msg, bp = self.get_breakpoint(bpnum)
        if not success:
            return False, msg
        self.delete_breakpoint(bp)
        return (True, '')

    def en_disable_all_breakpoints(self,  do_enable=True):
        "Enable or disable all breakpoints."
        bp_list = [bp for bp in self.bpbynumber if bp]
        bp_nums = []
        if do_enable:
            endis = 'en'
        else:
            endis = 'dis'
            pass
        if not bp_list:
            return "No breakpoints to %sable" % endis
        for bp in bp_list:
            bp.enabled = do_enable
            bp_nums.append(str(bp.number))
            pass
        return ("Breakpoints %sabled: %s" % (endis, ", ".join(bp_nums)))

    def en_disable_breakpoint_by_number(self, bpnum, do_enable=True):
        "Enable or disable a breakpoint given its breakpoint number."
        success, msg, bp = self.get_breakpoint(bpnum)
        if not success:
            return success, msg
        if do_enable:
            endis = 'en'
        else:
            endis = 'dis'
            pass
        if bp.enabled == do_enable:
            return (False, ('Breakpoint (%r) previously %sabled' %
                            (str(bpnum), endis,)))
        bp.enabled = do_enable
        return (True, '')

    def delete_breakpoints_by_lineno(self, filename, lineno):
        """Removes all breakpoints at a give filename and line number.
        Returns a list of breakpoints numbers deleted.
        """
        if (filename, lineno) not in self.bplist:
            return []
        breakpoints = self.bplist[(filename, lineno)]
        bpnums = [bp.number for bp in breakpoints]
        for bp in list(breakpoints):
            self.delete_breakpoint(bp)
        return bpnums

    def find_bp(self, filename, lineno, frame):
        """Determine which breakpoint for this file:line is to be acted upon.

        Called only if we know there is a bpt at this
        location.  Returns breakpoint that was triggered and a flag
        that indicates if it is ok to delete a temporary breakpoint.

        """
        possibles = self.bplist[filename, lineno]
        for i in range(0, len(possibles)):
            b = possibles[i]
            if not b.enabled:
                continue
            if not checkfuncname(b, frame):
                continue
            # Count every hit when bp is enabled
            b.hits += 1
            if not b.condition:
                # If unconditional, and ignoring, go on to next, else
                # break
                if b.ignore > 0:
                    b.ignore = b.ignore -1
                    continue
                else:
                    # breakpoint and marker that's ok to delete if
                    # temporary
                    return (b, True)
            else:
                # Conditional bp.
                # Ignore count applies only to those bpt hits where the
                # condition evaluates to true.
                try:
                    val = eval(b.condition, frame.f_globals, frame.f_locals)
                    if val:
                        if b.ignore > 0:
                            b.ignore = b.ignore -1
                            # continue
                        else:
                            return (b, True)
                    # else:
                    #   continue
                except:
                    # if eval fails, most conservative thing is to
                    # stop on breakpoint regardless of ignore count.
                    # Don't delete temporary, as another hint to user.
                    return (b, False)
                pass
            pass
        return (None, None)

    def last(self):
        return len(self.bpbynumber)-1

    def reset(self):
        """ A list of breakpoints by breakpoint number.  Each entry is
        None or an instance of Breakpoint.  Index 0 is unused, except
        for marking an effective break .... see effective(). """
        self.bpbynumber = [None]

        # A list of breakpoints indexed by (file, lineno) tuple
        self.bplist = {}
        self.fnlist  = {}

        return

    pass  # BreakpointManager


class Breakpoint:

    """Breakpoint class implements temporary breakpoints, ignore
    counts, disabling and (re)-enabling breakpoints and breakpoint
    conditionals.

    When matching an offset, if offset is None, the offset value is not
    considered in a match. If offset is -1, we are at a "call" event;
    the first offset in bytecode is 0.
    """

    def __init__(self, number, filename, line, temporary=False,
                 condition=None, funcname=None, offset=None):

        self.offset = offset
        self.condition = condition
        self.enabled   = True

        self.filename  = filename
        if filename: self.filename  = os.path.realpath(filename)

        # Needed if funcname is not None.
        self.func_first_executable_line = None
        self.funcname  = funcname

        # Number of time breakpoint has been hit
        self.hits      = 0

        # Number of times to ignore breakpoint before stopping
        self.ignore    = 0

        self.line      = line
        self.number    = number

        # Delete breakpoint after hitting it.
        self.temporary = temporary
        return

    def __str__(self):
        if self.temporary:
            disp = 'del  '
        else:
            disp = 'keep '
        if self.enabled:
            disp = disp + 'yes  '
        else:
            disp = disp + 'no   '
        if self.offset is None:
            offset_str = " any"
        else:
            offset_str = "%4d" % self.offset
        msg = '%-4dbreakpoint   %s %s at %s:%d' % (self.number, disp,
                                                    offset_str, self.filename, self.line)
        if self.condition:
            msg += '\n\tstop only if %s' % self.condition
        if self.ignore:
            msg += msg('\n\tignore next %d hits' % self.ignore)
        if self.hits:
            if (self.hits > 1): ss = 's'
            else: ss = ''
            msg +='\n\tbreakpoint already hit %d time%s' % self.hits, ss
        return msg

    def enable(self):
        self.enabled = True
        return self.enabled

    def disable(self):
        self.enabled = False
        return self.enabled

    def icon_char(self):
        """Return a one-character "icon" giving the state of the breakpoint
        't': temporary breakpoint
        'B': enabled breakpoint
        'b': disabled breakpoint
        """
        if self.temporary : return 't'
        elif self.enabled:  return 'B'
        else: return 'b'
        return

    pass  # end of Breakpoint class

def checkfuncname(b, frame):
    """Check whether we should break here because of `b.funcname`."""
    if not b.funcname:
        # Breakpoint was set via line number.
        if b.line != frame.f_lineno:
            # Breakpoint was set at a line with a def statement and the function
            # defined is called: don't break.
            return False
        return True

    # Breakpoint set via function name.

    if frame.f_code.co_name != b.funcname:
        # It's not a function call, but rather execution of def statement.
        return False

    # We are in the right frame.
    if not b.func_first_executable_line:
        # The function is entered for the 1st time.
        b.func_first_executable_line = frame.f_lineno

    if b.func_first_executable_line != frame.f_lineno:
        # But we are not at the first line number: don't break.
        return False
    return True

# Demo

if __name__=='__main__':
    bpmgr = BreakpointManager()
    print(bpmgr.last())
    bp = bpmgr.add_breakpoint('foo', 0, 5)
    print(bp.icon_char())
    print(bpmgr.last())
    print(repr(bp))
    print(str(bp))
    bp.disable()
    print(str(bp))
    for i in 10, 1:
        status, msg = bpmgr.delete_breakpoint_by_number(i)
        print("Delete breakpoint %s: %s %s" % (i, status, msg,))
    import inspect
    frame = inspect.currentframe()
    print("Stop at bp: %s" % checkfuncname(bp, frame))

    def foo(bp, bpmgr):
        frame = inspect.currentframe()
        print("Stop at bp2: %s" % checkfuncname(bp, frame))
        # frame.f_lineno is constantly updated. So adjust for the
        # line difference between the add_breakpoint and the check.
        bp3 = bpmgr.add_breakpoint('foo', 0, frame.f_lineno+1)
        print("Stop at bp3: %s" % checkfuncname(bp3, frame))
        return

    bp2 = bpmgr.add_breakpoint(None, None, False, None, 'foo')
    foo(bp2, bpmgr)
    bp3 = bpmgr.add_breakpoint('foo', 5, 2, temporary=True)
    print(bp3.icon_char())
    print(bpmgr.bpnumbers())

    bp = bpmgr.add_breakpoint('bar', 10, 3)
    filename = bp.filename
    for i in range(3):
        bp = bpmgr.add_breakpoint('bar', 2, 6)
    print(bpmgr.delete_breakpoints_by_lineno(filename, 6))
    print(bpmgr.delete_breakpoints_by_lineno(filename, 6))
    print(bpmgr.bpnumbers())



#!/usr/bin/env python3

"""
distutils setup (setup.py).

This is just boilerplate code, since we do like to try to keep data separate
from code as much as possible. The customizable information really comes
from file __pkginfo__.py.
"""

from setuptools import setup

setup(
       test_suite         = 'nose.collector',
       setup_requires     = ['nose>=1.0'],
       name               = 'testing_test',
       )



# -*- coding: utf-8 -*-
"""Deparsing Routines"""

import sys, tempfile
from io import StringIO
from hashlib import sha1
from uncompyle6.semantics.linemap import code_deparse_with_map
from uncompyle6.semantics.fragments import deparsed_find, code_deparse
import pyficache

# FIXME remap filename to a short name.

deparse_cache = {}


def deparse_and_cache(co, errmsg_fn, tempdir=None):
    # co = proc_obj.curframe.f_code
    out = StringIO()
    deparsed = deparse_cache.get(co, None)
    if not deparsed or not hasattr(deparsed, "source_linemap"):
        try:
            deparsed = code_deparse_with_map(co, out)
        except:
            errmsg_fn(str(sys.exc_info()[0]))
            errmsg_fn("error in deparsing code: %s" % co.co_filename)
            return None, None

        deparse_cache[co] = deparsed

    text = out.getvalue()
    linemap = [
        (line_no, deparsed.source_linemap[line_no])
        for line_no in sorted(deparsed.source_linemap.keys())
    ]

    # FIXME: DRY code with version in cmdproc.py print_location

    name_for_code = sha1(co.co_code).hexdigest()[:6]
    prefix = "deparsed-"
    fd = tempfile.NamedTemporaryFile(
        suffix=".py", prefix=prefix, dir=tempdir, delete=False
    )
    with fd:
        fd.write(text.encode("utf-8"))
        map_line = "\n\n# %s" % linemap
        fd.write(map_line.encode("utf-8"))
        remapped_file = fd.name
    fd.close()
    # FIXME remap filename to a short name.
    pyficache.remap_file_lines(name_for_code, remapped_file, linemap)
    return remapped_file, name_for_code


def deparse_offset(co, name, last_i, errmsg_fn):
    nodeInfo = None
    deparsed = deparse_cache.get(co, None)
    if not deparsed or not hasattr(deparsed, "offsets"):
        out = StringIO()
        try:
            # FIXME: cache co
            deparsed = code_deparse(co, out)
        except:
            print(sys.exc_info()[1])
            if errmsg_fn:
                errmsg_fn(sys.exc_info()[1])
                errmsg_fn("error in deparsing code")
        deparse_cache[co] = deparsed
    try:
        nodeInfo = deparsed_find((name, last_i), deparsed, co)
    except:
        if errmsg_fn:
            errmsg_fn(sys.exc_info()[1])
            errmsg_fn("error in deparsing code at offset %d" % last_i)

    if not nodeInfo:
        nodeInfo = deparsed_find((name, last_i), deparsed, co)
    return deparsed, nodeInfo


# Demo it
if __name__ == "__main__":
    import inspect

    def msg(msg_str):
        print(msg_str)
        return

    def errmsg(msg_str):
        msg("*** " + msg_str)
        return

    curframe = inspect.currentframe()
    line_no = curframe.f_lineno
    mapped_name, name_for_code = deparse_and_cache(curframe.f_code, errmsg)
    print(pyficache.getline(mapped_name, 7))
    # mapped_name, name_for_code = deparse_offset(curframe.f_code,
    #                                             curframe.f_code.co_name,
    #                                             curframe.f_lasti, errmsg)
    # print(pyficache.getline(mapped_name, 7))



# -*- coding: utf-8 -*-
#  Copyright (C) 2009-2010, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.complete import complete_bpnumber


class ConditionCommand(DebuggerCommand):
    """**condition** *bp_number* *condition*

*bp_number* is a breakpoint number. *condition* is an expression which
must evaluate to *True* before the breakpoint is honored.  If *condition*
is absent, any existing condition is removed; i.e., the breakpoint is
made unconditional.

Examples:
---------

   condition 5 x > 10  # Breakpoint 5 now has condition x > 10
   condition 5         # Remove above condition

See also:
---------

`break`, `tbreak`."""

    aliases = ("cond",)
    short_help = "Specify breakpoint number N to break only if COND is True"

    DebuggerCommand.setup(locals(), category="breakpoints", min_args=1)

    complete = complete_bpnumber

    def run(self, args):
        success, msg, bp = self.core.bpmgr.get_breakpoint(int(args[1]))
        if not success:
            self.errmsg(msg)
            return
        if len(args) > 2:
            condition = " ".join(args[2:])
        else:
            condition = None
            self.msg("Breakpoint %d is now unconditional." % bp.number)
            pass
        bp.condition = condition
        return


if __name__ == "__main__":
    import sys
    from trepan import debugger as Mdebugger

    Mbreak = __import__("trepan.processor.command.break", None, None, ["*"])
    d = Mdebugger.Trepan()
    brkcmd = Mbreak.BreakCommand(d.core.processor)
    command = ConditionCommand(d.core.processor)
    command.proc.frame = sys._getframe()
    command.proc.setup()

    command.run(["condition", "1"])
    brkcmd.run(["break"])
    command.run(["condition", "1"])
    command.run(["condition", "1", "x", ">", "10"])
    command.run(["condition", "1"])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2021 Rocky Bernstein
#

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand


class ShowTempdir(DebuggerSubcommand):
    """**show tempdir**

Show the temporary directory usind in decompiled python files.

See also:
--------

`set tempdir`
"""

    in_list = True
    min_abbrev = len("temp")
    min_args = 0
    max_args = 0
    short_help = "Set a directory for storing decompiled Python"

    def run(self, args):
        tempdir = self.debugger.settings.get(self.name, None)
        if tempdir:
            self.msg("tempdir is %s." % tempdir)
        else:
            self.msg("tempdir not set; Python default is used.")
        return

    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    sub = Mhelper.demo_run(ShowTempdir)
    d = sub.proc.debugger
    sub.run(["show"])
    pass



# -*- coding: utf-8 -*-
""" Mock classes used for testing which in turn really come from 
trepan.processor.command.mock """

from trepan.lib import default
from trepan.processor.command import mock

dbg_setup = mock.dbg_setup



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2010, 2013-2015, 2020-2021 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


def option_set(options, value, default_options):
    if not options or value not in options:
        return default_options.get(value)
    else:
        return options.get(value)
    return None  # Not reached


def bool2YN(b) -> bool:
    return "Y" if b else "N"


def wrapped_lines(msg_part1, msg_part2, width):
    if len(msg_part1) + len(msg_part2) + 1 > width:
        return msg_part1 + "\n\t" + msg_part2
    else:
        return msg_part1 + " " + msg_part2
    return  # Not reached


def pretty_modfunc_name(s) -> str:
    if s == "<module>":
        # FIXME:
        # Pick replace with something more custom to modname?
        return str(s)
    else:
        return str(s) + "()"


import os
from glob import glob


def pyfiles(callername, level=2):
    "All python files caller's dir without the path and trailing .py"
    d = os.path.dirname(callername)
    # Get the name of our directory.
    # A glob pattern that will get all *.py files but not __init__.py
    glob(os.path.join(d, "[a-zA-Z]*.py"))
    py_files = glob(os.path.join(d, "[a-zA-Z]*.py"))
    return [os.path.basename(filename[0:-3]) for filename in py_files]


# Demo it
if __name__ == "__main__":
    TEST_OPTS = {"a": True, "b": 5, "c": None}
    get_option = lambda key: option_set(opts, key, TEST_OPTS)
    opts = {"d": 6, "a": False}
    for opt in ["a", "b", "c", "d"]:
        print(opt, get_option(opt))
        pass
    for b in [True, False]:
        print(bool2YN(b))
    pass

    print(wrapped_lines("hi", "there", 80))
    print(wrapped_lines("hi", "there", 5))
    print(pyfiles(__file__))
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009-2010, 2013, 2015, 2017, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import code, os.path as osp, sys

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.interfaces.server import ServerInterface


class PythonCommand(DebuggerCommand):
    """**python** [**-d**]

Run Python as a command subshell. The *sys.ps1* prompt will be set to
`trepan3 >>> `.

If *-d* is passed, you can access debugger state via local variable *debugger*.

To issue a debugger command use function *dbgr()*. For example:

  dbgr('info program')

See also:
---------

`ipython`, `bpython`
"""

    aliases = ("py", "shell")
    short_help = "Run Python as a command subshell"

    DebuggerCommand.setup(locals(), category="data", max_args=1)

    def dbgr(self, string):
        """Invoke a debugger command from inside a python shell called inside
        the debugger.
        """
        self.proc.cmd_queue.append(string)
        self.proc.process_command()
        return

    def run(self, args):
        # See if python's code module is around

        # Python does it's own history thing.
        # Make sure it doesn't damage ours.
        intf = self.debugger.intf[-1]
        if isinstance(intf, ServerInterface):
            self.errmsg("Can't run an interactive shell on a remote session")
            return

        have_line_edit = self.debugger.intf[-1].input.line_edit
        if have_line_edit:
            try:
                self.proc.write_history_file()
            except IOError:
                pass
            pass

        banner_tmpl = """trepan3 python shell%s
Use dbgr(*string*) to issue debugger command: *string*"""

        debug = len(args) > 1 and args[1] == "-d"
        if debug:
            banner_tmpl += "\nVariable 'debugger' contains a trepan" "debugger object."
            pass

        my_locals = {}
        my_globals = None
        proc = self.proc
        if proc.curframe:
            my_globals = proc.curframe.f_globals
            if proc.curframe.f_locals:
                my_locals = proc.curframe.f_locals
                pass
            pass

        # Give python and the user a way to get access to the debugger.
        if debug:
            my_locals["debugger"] = self.debugger
        my_locals["dbgr"] = self.dbgr

        # Change from debugger completion to python completion
        try:
            import readline
        except ImportError:
            pass
        else:
            readline.parse_and_bind("tab: complete")

        sys.ps1 = "trepan3k >>> "
        old_sys_excepthook = sys.excepthook
        try:
            sys.excepthook = None
            if len(my_locals):
                interact(
                    banner=(banner_tmpl % " with locals"),
                    my_locals=my_locals,
                    my_globals=my_globals,
                )
            else:
                interact(banner=(banner_tmpl % ""))
                pass
        finally:
            sys.excepthook = old_sys_excepthook

        # restore completion and our history if we can do so.
        if hasattr(proc.intf[-1], "complete"):
            try:
                from readline import set_completer, parse_and_bind

                parse_and_bind("tab: complete")
                set_completer(proc.intf[-1].complete)
            except ImportError:
                pass
            pass

        if have_line_edit:
            proc.read_history_file()
            pass
        return

    pass


# Monkey-patched from code.py
# FIXME: get changes into Python.
def interact(banner=None, readfunc=None, my_locals=None, my_globals=None):
    """Almost a copy of code.interact
    Closely emulate the interactive Python interpreter.

    This is a backwards compatible interface to the InteractiveConsole
    class.  When readfunc is not specified, it attempts to import the
    readline module to enable GNU readline if it is available.

    Arguments (all optional, all default to None):

    banner -- passed to InteractiveConsole.interact()
    readfunc -- if not None, replaces InteractiveConsole.raw_input()
    local -- passed to InteractiveInterpreter.__init__()

    """

    def console_runcode(code_obj):
        runcode(console, code_obj)

    console = code.InteractiveConsole(my_locals, filename="<trepan>")
    console.runcode = console_runcode
    setattr(console, "globals", my_globals)
    if readfunc is not None:
        console.raw_input = readfunc
    else:
        try:
            import readline
        except ImportError:
            pass
    console.interact(banner)
    pass


# Also monkey-patched from code.py
# FIXME: get changes into Python.
def runcode(obj, code_obj):
    """Execute a code object.

    When an exception occurs, self.showtraceback() is called to
    display a traceback.  All exceptions are caught except
    SystemExit, which is reraised.

    A note about KeyboardInterrupt: this exception may occur
    elsewhere in this code, and may not always be caught.  The
    caller should be prepared to deal with it.

    """
    try:
        exec(code_obj, obj.locals, obj.globals)
    except SystemExit:
        raise
    except:
        info = sys.exc_info()
        print("%s; %s" % (info[0], info[1]))
    else:
        pass
    return


if __name__ == "__main__":
    from trepan.debugger import Trepan

    d = Trepan()
    command = PythonCommand(d.core.processor)
    command.proc.frame = sys._getframe()
    command.proc.setup()
    if len(sys.argv) > 1:
        print("Type Python commands and exit to quit.")
        print(sys.argv[1])
        if sys.argv[1] == "-d":
            print(command.run(["python", "-d"]))
        else:
            print(command.run(["python"]))
            pass
        pass
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand


class ShowAsmFmt(DebuggerSubcommand):
    """**show asmfmt**

Show the disassembly format style used in the `disassemble` cmmand.

See also:
---------

`set asmfmt`"""

    min_abbrev = len("asmf")
    short_help = "Show assembly format style"
    pass


    def run(self, args):
        if len(args) != 0:
            self.errmsg("Expecting no args")
            return

        style = self.debugger.settings[self.name]
        if style:
            self.msg("Assembly format style is %s" % style)
        else:
            self.msg("Assembly format style not set")
        return

    pass
if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    sub = Mhelper.demo_run(ShowAsmFmt, [])
    d = sub.proc.debugger
    sub.run(["invalid arg"])
    pass



#!/usr/bin/env python3
'Unit test for trepan.ionout.fifo*'
import os
if hasattr(os, 'mkfifo'):

    import unittest

    from trepan.inout import fifoserver as Mserver, fifoclient as Mclient

    class TestFIFO(unittest.TestCase):
        """Tests FIFOServer and FIFOClient"""

        def test_client_server(self):
            Mserver.FIFOServer(opts={'open': True})
            Mclient.FIFOClient(opts={'open': os.getpid()})
            self.assertTrue(True, 'FIXME: need to add a test here.')
            # FIXME need to use threading or forking
    #         for line in ['one', 'two', 'three']:
    #             server.writeline(line)
    #             self.assertEqual(line, client.readline())
    #             pass
    #         for line in ['four', 'five', 'six']:
    #             client.writeline(line)
    #             self.assertEqual(line, server.readline())
    #             pass
            return

    if __name__ == '__main__':
        unittest.main()
        pass
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013, 2015, 2017 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.cmdbreak import parse_break_cmd, set_break
from trepan.processor.complete import complete_break_linenumber


class TempBreakCommand(DebuggerCommand):
    """**tbreak** [*location*] [**if** *condition*]

Sets a temporary breakpoint, i.e. one that is removed the after
the first time it is encountered.

See the help for location for what can be specified there.

Without arguments or an empty *location*, the temporary breakpoint is
set at the current stopped location.

If the word `if` is given after *location*, subsequent arguments given
a boolean condition which must evaluate to True before the breakpoint
is honored.

Examples:
---------

   tbreak                # Break where we are current stopped at
   tbreak if i < j       # Break at current line if i < j
   tbreak 10             # Break on line 10 of the file we are
                         # currently stopped at
   tbreak os.path.join() # Break in function os.path.join
   tbreak x[i].fn() if x # break in function specified by x[i].fn
                         # if x is set
   tbreak os.path:45     # Break on line 45 file holding module os.path
   tbreak myfile.py:45   # Break on line 45 of myfile.py
   break '''c:\\foo.bat''':1"  # One way to specify a Windows file name,
   break '''/My Docs/foo.py''':1"  # One way to specify path with blanks in it

See also:
---------

`break`, `condition` and `help syntax location`.
"""

    aliases = ("tb", "tbreak!", "tb!")
    min_args      = 0
    short_help    = 'Set temporary breakpoint at specified line or function'

    DebuggerCommand.setup(locals(), category="breakpoints", need_stack=True)

    complete = complete_break_linenumber

    def run(self, args):
        func, filename, lineno, condition, offset = parse_break_cmd(self.proc, args)
        if not (func == None and filename == None):
            set_break(self, func, filename, lineno, condition,
                      True, args)
        return

if __name__ == '__main__':
    from trepan.debugger import Trepan

    d = Trepan()
    command = TempBreakCommand(d.core.processor)
    command.proc.frame = sys._getframe()
    command.proc.setup()

    def doit(args):
        command.proc.current_command = " ".join(args)
        print(parse_break_cmd(command.proc, args))

    d = Trepan()

    print(doit(["10"]))
    print(doit([__file__ + ':10']))

    def foo():
        return 'bar'
    print(doit(['foo']))
    print(doit(['os.path']))
    print(doit(['os.path', '5+1']))
    print(doit(['os.path.join']))
    print(doit(['if', 'True']))
    print(doit(['foo', 'if', 'True']))
    print(doit(['os.path:10', 'if', 'True']))
    command.run(['tbreak'])
    command.run(['tbreak', 'command.run'])
    command.run(['tbreak', '10'])
    command.run(['tbreak', __file__ + ':10'])
    command.run(['tbreak', 'foo'])
    pass



# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class ShowDbgTrepan(Mbase_subcmd.DebuggerShowBoolSubcommand):
    """**show dbg_trepan**

Show debugging the debugger

See also:
---------

`set dbg_trepan`"""
    min_abbrev = 4  # Min 'show trep"
    short_help = 'Show debugging the debugger'
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2008-2009, 2013, 2015, 2021 Rocky Bernstein <rocky@gnu.org>
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
|Downloads| |Build Status|

Abstract
========
A gdb-like debugger for Python3.


A command-line interface (CLI) is provided.

See the `Tutorial <https://github.com/rocky/python2-trepan/wiki/Tutorial>`_ for how to use.

Features
========

There's a lot of cool stuff here that's not in *pydb* or the stock
Python debugger *pdb*.

Source-code Syntax Colorization
-------------------------------

Terminal source code is colorized and we make use of terminal bold and
emphasized text in debugger output and help text. Of course, you can
also turn this off.

Smart Eval
----------

If you want to evaluate the current source line before it is run in
the code, use `eval`. To evaluate text of a common fragment of line,
such as the expression part of an *if* statement, you can do that with
`eval?`. See the help for `eval` for more information.

Better stepping granularity
---------------------------

Sometimes you want small steps, and sometimes large stepping.

This fundamental issue is handled in a couple ways:

Step Granularity
----------------

There are now `step` *event* and `next` *event* commands with aliases
to `s+`, `s>` and so on. The plus-suffixed commands force a different
line on a subsequent stop, the dash-suffixed commands don't.  Suffixes
`>`, `<`, and `!` specify `call`, `return` and `exception` events
respectively. And without a suffix you get the default; this is set by
the `set different` command.

Event Filtering and Tracing
---------------------------

By default the debugger stops at every event: `call`, `return`,
`line`, `exception`, `c-call`, `c-exception`. If you just want to stop
at `line` events (which is largely what you happens in _pdb_) you
can. If however you just want to stop at calls and returns, that's
possible too. Or pick some combination.

In conjunction with handling *all* events by default, the event status
is shown when stopped. The reason for stopping is also available via
`info program`.

Event Tracing of Calls and Returns
----------------------------------

I'm not sure why this was not done before. Probably because of the
lack of the ability to set and move by different granularities,
tracing calls and returns lead to too many uninteresting stops (such
as at the same place you just were at). Also, stopping on function
definitions probably also added to this tedium.

Because we're really handling return events, we can show you the
return value. (_pdb_ has an \"undocumented\" _retval_ command that
doesn't seem to work.)

Debugger Macros via Python Lambda expressions
---------------------------------------------

In *gdb*, there is a *macro* debugger command to extend debugger
commands. However Python has its own rich programming language so it
seems silly to recreate the macro language that is in *gdb*. Simpler
and more powerful is just to use Python here. A debugger macro here is
just a lambda expression which returns a string or a list of
strings. Each string returned should be a debugger command.

We also have _aliases_ for the extremely simple situation where you
want to give an alias to an existing debugger command. But beware:
some commands, like `step` inspect command suffixes and change their
behavior accordingly.

We also envision a number of other ways to allow extension of this
debugger either through additional modules, or user-supplied debugger
command directories.

If what you were looking for in macros was more front-end control over
the debugger, then consider using the experimental (and not finished)
Bullwinkle protocol.

Byte-code Instruction Introspection
-----------------------------------

We do more in the way of looking at the byte codes to give better
information. Through this we can provide:

* a *skip* command. It is like the *jump* command, but you don't have to deal
with line numbers.
* disassembly of code fragments. You can now disassemble relative to
  the stack frames

  you are currently stopped at.
* Better interpretation of where you are when inside execfile or exec.
  (But really though this is probably a Python compiler misfeature.)
* Check that breakpoints are set only where they make sense.
* A more accurate determination of if you are at a function-defining *def*
  statement (because the caller instruction contains *MAKE_FUNCTION*.)

Debugger Command Arguments can be Variables and Expressions
===========================================================

Commands that take integer arguments like *up*, *list* or
*disassemble* allow you to use a Python expression which may include
local or global variables that evaluates to an integer. This
eliminates the need in *gdb* for special \"dollar\" debugger
variables. (Note however because of _shlex_ parsing expressions can't
have embedded blanks.)

Modularity
==========

The Debugger plays nice with other trace hooks. You can have several
debugger objects.

Many of the things listed below doesn't directly effect end-users, but
it does eventually by way of more robust and featureful code. And
keeping developers happy is a good thing.(TM)

* Commands and subcommands are individual classes now, not methods in a class.
  This means they now have properties like the context in which they can
  be run,  minimum abbreviation name or alias names.
  To add a new command you basically add a file in a directory.

* I/O is it's own layer. This simplifies interactive readline behavior
  from reading commands over a TCP socket.
* An interface is it's own layer. Local debugging, remote debugging,
  running debugger commands from a file (`source`) are different interfaces.
  This means, for example, that we are able to give better error reporting
  if a debugger command file has an error.
* There is an experimental Python-friendly interface for front-ends
* more testable. Much more unit and functional tests. More of _pydb_'s
  integration test will eventually be added.

Copyright (C) 2008-2009, 2013-2016, 2021 Rocky Bernstein <rocky@gnu.org>

.. |Downloads| image:: https://pypip.in/download/trepan/badge.svg
.. |Build Status| image:: https://travis-ci.org/rocky/python3-trepan.svg
"""
__docformat__ = "restructuredtext"

from trepan.version import __version__



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2012-2017, 2020-2021 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import inspect, linecache, pyficache, sys
import os.path as osp

# Our local modules
from pygments.console import colorize

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.cmdlist import parse_list_cmd
from trepan.lib.deparse import deparse_cache


class ListCommand(DebuggerCommand):
    """**list** [ *range* ]

**list**  **+** | **-** | **.**

List source code. See `help syntax range` for what can go in a list range.

Without arguments, print lines starting from where the last list left off
since the last entry to the debugger. We start off at the location indicated
by the current stack.

in addition you can also use:

  - a '.' for the location of the current frame
  - a '-' for the lines before the last list
  - a '+' for the lines after the last list

Examples:
--------

    list 5               # List starting from line 5 of current file
    list 5 ,             # Same as above.
    list , 5             # list listsize lines before and up to 5
    list foo.py:5        # List starting from line 5 of file foo.py
    list foo()           # List starting from function foo
    list os.path:5       # List starting from line 5 of module os.path
    list os.path:5, 6    # list lines 5 and 6 of os.path
    list os.path:5, +1   # Same as above. +1 is an offset
    list os.path:5, 1    # Same as above, since 1 < 5.
    list os.path:5, +6   # list lines 5-11
    list os.path.join()  # List lines centered around the os.join.path function.
    list .               # List lines centered from where we currently are stopped
    list -               # List lines previous to those just shown
    list                 # List continuing from where we last left off

See also:
---------

`set listize` or `show listsize` to see or set the value; `help syntax location`
for the specification of a location and `help syntax range` for the specification
of a range.
"""

    aliases = ("l",)
    short_help = "List source code"

    DebuggerCommand.setup(locals(), category="files", max_args=3)

    def run(self, args):
        proc = self.proc
        dbg_obj = self.core.debugger
        listsize = dbg_obj.settings["listsize"]
        filename, first, last = parse_list_cmd(proc, args, listsize)
        curframe = proc.curframe
        if filename is None:
            return
        filename = pyficache.unmap_file(pyficache.resolve_name_to_path(filename))

        # We now have range information. Do the listing.
        max_line = pyficache.size(filename)
        if max_line is None:
            bytecode = curframe.f_code

            if bytecode not in deparse_cache.keys():
                self.errmsg(
                    'No file %s found; using "deparse" command instead to show source'
                    % filename
                )
            proc.commands["deparse"].run(["deparse"])
            return

        canonic_filename = osp.realpath(osp.normcase(filename))

        if first > max_line:
            self.errmsg(
                'Bad start line %d - file "%s" has only %d lines'
                % (first, filename, max_line)
            )
            return

        if last > max_line:
            self.msg("End position changed to last line %d " % max_line)
            last = max_line

        bplist = self.core.bpmgr.bplist
        opts = {
            "reload_on_change": self.settings["reload"],
            "output": self.settings["highlight"],
            "strip_nl": False,
        }

        if "style" in self.settings:
            opts["style"] = self.settings["style"]

        if first <= 0:
            first = 1
        try:
            for lineno in range(first, last + 1):
                line = pyficache.getline(filename, lineno, opts)
                if line is None:
                    line = linecache.getline(filename, lineno, proc.frame.f_globals)
                    pass
                if line is None:
                    self.msg("[EOF]")
                    break
                else:
                    line = line.rstrip("\n")
                    s = proc._saferepr(lineno).rjust(3)
                    if len(s) < 5:
                        s += " "
                    if (canonic_filename, lineno,) in list(bplist.keys()):
                        bp = bplist[(canonic_filename, lineno,)][0]
                        a_pad = "%02d" % bp.number
                        s += bp.icon_char()
                    else:
                        s += " "
                        a_pad = "  "
                        pass
                    if curframe and lineno == inspect.getlineno(curframe):
                        s += "->"
                        if "plain" != self.settings["highlight"]:
                            s = colorize("bold", s)
                    else:
                        s += a_pad
                        pass
                    self.msg(s + "\t" + line)
                    proc.list_lineno = lineno
                    pass
                pass
            pass
        except KeyboardInterrupt:
            pass
        return False


if __name__ == "__main__":

    # FIXME: make sure the below is in a unit test
    def doit(cmd, args):
        proc = cmd.proc
        proc.current_command = " ".join(args)
        cmd.run(args)

    from trepan.processor.command import mock as Mmock
    from trepan.processor.cmdproc import CommandProcessor

    d = Mmock.MockDebugger()
    cmdproc = CommandProcessor(d.core)
    cmdproc.frame = sys._getframe()
    cmdproc.setup()
    lcmd = ListCommand(cmdproc)

    print("--" * 10)
    # doit(lcmd, ['list'])

    # doit(lcmd, ['list', __file__ + ':10'])
    # print('--' * 10)

    # doit(lcmd, ['list', 'os:10'])
    # print('--' * 10)

    # doit(lcmd, ['list', '.'])
    # print('--' * 10)

    # doit(lcmd, ['list', '10'])
    # print('--' * 10)

    # doit(lcmd, ['list', '1000'])

    def foo():
        return "bar"

    # doit(lcmd, ['list', 'foo()'])
    # print('--' * 10)
    # doit(lcmd, ['list'])
    # print('--' * 10)
    # doit(lcmd, ['list', '-'])
    # doit(lcmd, ['list', '-'])
    # doit(lcmd, ['list', '+'])
    # doit(lcmd, ['list', '+'])
    doit(lcmd, ["list", "40,", "60"])
    # doit(lcmd, ['list', '20', '5'])

    # doit(lcmd, ['list', 'os.path'])
    # print('--' * 10)
    # doit(lcmd, ['list', 'os.path', '15'])
    # print('--' * 10)
    # doit(lcmd, ['list', 'os.path', '30', '3'])
    # print('--' * 10)
    # doit(lcmd, ['list', 'os.path', '40', '50'])
    # print('--' * 10)

    # doit(lcmd, ['list', os.path.abspath(__file__)+':3', '4'])
    # print('--' * 10)
    # doit(lcmd, ['list', os.path.abspath(__file__)+':3', '12-10'])
    # doit(lcmd, ['list', 'os.path:5'])
    pass



# -*- coding: utf-8 -*-
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os, threading

# Our local modules
from trepan.bwprocessor.command import base_cmd as Mbase_cmd
from trepan import exception as Mexcept


class QuitCommand(Mbase_cmd.DebuggerCommand):
    """Gently exit the debugger and debugged program.

**Input Fields:**

   { command     => 'quit',
   }

The program being debugged is exited raising a *DebuggerQuit* exception.

**Output Fields: **

   { name      => 'status',
     event     => 'terminated',
     [errmsg   => <error-message-array>]
     [msg      => <message-text array>]
  }
"""

    name          = os.path.basename(__file__).split('.')[0]
    need_stack    = False

    def nothread_quit(self, arg):
        """ quit command when there's just one thread. """

        self.debugger.core.stop()
        self.debugger.core.execution_status = 'Quit command'
        self.proc.response['event'] = 'terminated'
        self.proc.response['name']  = 'status'
        self.proc.intf[-1].msg(self.proc.response)
        raise Mexcept.DebuggerQuit

    def threaded_quit(self, arg):
        """ quit command when several threads are involved. """
        self.msg("Quit for threading not fully done yet. Try 'kill'.")
        return False

    def run(self, cmd_hash):
        threading_list = threading.enumerate()
        if (len(threading_list) == 1 and
            threading_list[0].getName() == 'MainThread'):
            # We just have a main thread so that's safe to quit
            return self.nothread_quit(cmd_hash)
        else:
            return self.threaded_quit(cmd_hash)
        pass

if __name__ == '__main__':
    from trepan.bwprocessor.command import mock
    d, cp = mock.dbg_setup()
    command = QuitCommand(cp)
    try:
        command.run(['quit'])
    except Mexcept.DebuggerQuit:
        print("A got 'quit' a exception. Ok, be that way - I'm going home.")
        pass

    class MyThread(threading.Thread):
        def run(self):
            command.run(['quit'])
            return
        pass

    t = MyThread()
    t.start()
    t.join()
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import tracer

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.cmdfns import want_different_line


class StepCommand(DebuggerCommand):
    """**step**[**+**|**-**|**<**|**>**|**!**] [*event*...] [*count*]

Execute the current line, stopping at the next event.

With an integer argument, step that many times.

*event* is list of an event name which is one of: `call`,
`return`, `line`, `exception` `c-call`, `c-return` or `c-exception`.
If specified, only those stepping events will be considered. If no
list of event names is given, then any event triggers a stop when the
count is 0.

There is however another way to specify a *single* event, by
suffixing one of the symbols `<`, `>`, or `!` after the command or on
an alias of that.  A suffix of `+` on a command or an alias forces a
move to another line, while a suffix of `-` disables this requirement.
A suffix of `>` will continue until the next call. (`finish` will run
run until the return for that call.)

If no suffix is given, the debugger setting `different-line`
determines this behavior.

Examples:
---------

  step        # step 1 event, *any* event
  step 1      # same as above
  step 5/5+0  # same as above
  step line   # step only line events
  step call   # step only call events
  step>       # same as above
  step call line # Step line *and* call events

Related and similar is the `next` command.

See also:
---------

`next`, `skip`, `jump` (there's no `hop` yet), `continue`, `return` and
`finish` for other ways to progress execution.
"""

    aliases = (
        "step+",
        "step-",
        "step>",
        "step<",
        "step!",
        "s",
        "s+",
        "s-",
        "s<",
        "s>",
        "s!",
    )
    execution_set = ["Running"]
    short_help = "Step program (possibly entering called functions)"

    DebuggerCommand.setup(locals(), category="running", max_args=1, need_stack=True)

    def run(self, args):
        step_events = []
        if args[0][-1] == ">":
            step_events = ["call"]
        elif args[0][-1] == "<":
            step_events = ["return"]
        elif args[0][-1] == "!":
            step_events = ["exception"]
            pass
        if len(args) <= 1:
            self.proc.debugger.core.step_ignore = 0
        else:
            pos = 1
            while pos < len(args):
                arg = args[pos]
                if arg in tracer.ALL_EVENT_NAMES:
                    step_events.append(arg)
                else:
                    break
                pos += 1
                pass
            if pos == len(args) - 1:
                self.core.step_ignore = self.proc.get_int(
                    args[pos], default=1, cmdname="step"
                )
                if self.core.step_ignore is None:
                    return False
                # 0 means stop now or step 1, so we subtract 1.
                self.core.step_ignore -= 1
                pass
            elif pos != len(args):
                self.errmsg("Invalid additional parameters %s" % " ".join(args[pos]))
                return False
            pass

        if [] == step_events:
            self.core.step_events = None
        else:
            self.core.step_events = step_events
            pass

        self.core.different_line = want_different_line(
            args[0], self.settings["different"]
        )
        self.core.stop_level = None
        self.core.last_frame = None
        self.core.stop_on_finish = False
        self.proc.continue_running = True  # Break out of command read loop
        return True

    pass


if __name__ == "__main__":
    from mock import MockDebugger

    d = MockDebugger()
    cmd = StepCommand(d.core.processor)
    for c in (["s", "5"], ["step", "1+2"], ["s", "foo"]):
        d.core.step_ignore = 0
        cmd.proc.continue_running = False
        result = cmd.run(c)
        print("Execute result: %s" % result)
        print("step_ignore %s" % repr(d.core.step_ignore))
        print("continue_running: %s" % cmd.proc.continue_running)
        pass
    for c in (["s"], ["step+"], ["s-"], ["s!"], ["s>"], ["s<"]):
        d.core.step_ignore = 0
        cmd.continue_running = False
        result = cmd.run(c)
        print("different line %s:" % c[0], cmd.core.different_line)
        pass
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009-2010, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand
from pprint import pformat

class InfoReturn(DebuggerSubcommand):
    """return value

Show the value that is to be returned from a function.  This command
is useful after a 'finish' command or stepping just after a 'return'
statement."""

    min_abbrev    = 1
    need_stack    = True
    short_help    = 'Show function return value'

    def run(self, args):
        # pdb checks to see if __return__ is the frame's f_locals which doesn't work
        # at least on any Python I am aware of back to 2.4.
        # Testing on the event however does work.
        if self.proc.event in ['return', 'exception']:
            val = self.proc.event_arg
            formatted_val = pformat(val)
            self.msg("return value (type %s):\n\t%s" % (type(val), formatted_val))
        else:
            self.errmsg("Must be in a 'return' or 'exception' event "
                        "rather than a %s event." % self.proc.event)
            pass
        return

if __name__ == '__main__':
    from trepan.processor.command import mock, info as Minfo
    d, cp = mock.dbg_setup()
    i = Minfo.InfoCommand(cp)
    sub = InfoReturn(i)
    print(sub.run([]))
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand

class PythonCommand(DebuggerCommand):
    """**bpython** [**-d**]

Run bpython as a command subshell. The *sys.ps1* prompt will be set to
`trepan2 >>> `.

If *-d* is passed, you can access debugger state via local variable *debugger*.

To issue a debugger command use function *dbgr()*. For example:

  dbgr('info program')
"""

    short_help    = 'Run bpython as a command subshell'

    DebuggerCommand.setup(locals(), category="support", max_args=1)

    def dbgr(self, string):
        '''Invoke a debugger command from inside a python shell called inside
        the debugger.
        '''
        print('')
        self.proc.cmd_queue.append(string)
        self.proc.process_command()
        return

    def run(self, args):

        try:
            from bpython.curtsies import main as main_bpython
        except ImportError:
            self.errmsg("bpython needs to be installed to run this command")
            return

        # bpython does it's own history thing.
        # Make sure it doesn't damage ours.
        have_line_edit = self.debugger.intf[-1].input.line_edit
        if have_line_edit:
            try:
                self.proc.write_history_file()
            except IOError:
                pass
            pass

        banner_tmpl='''trepan2 python shell%s
Use dbgr(*string*) to issue debugger command: *string*'''

        debug = len(args) > 1 and args[1] == '-d'
        if debug:
            banner_tmpl += ("\nVariable 'debugger' contains a trepan" +
                            "debugger object.")
            pass
        my_locals  = {}
        # my_globals = None
        if self.proc.curframe:
            # my_globals = self.proc.curframe.f_globals
            if self.proc.curframe.f_locals:
                my_locals = self.proc.curframe.f_locals
                pass
            pass

        # Give python and the user a way to get access to the debugger.
        if debug: my_locals['debugger'] = self.debugger
        my_locals['dbgr'] = self.dbgr

        if len(my_locals):
            banner=(banner_tmpl % ' with locals')
        else:
            banner=(banner_tmpl % '')
            pass

        sys.ps1 = 'trepan2 >>> '
        print(banner)
        try:
            main_bpython([], my_locals)
        except SystemExit:
            pass

        # restore completion and our history if we can do so.
        if hasattr(self.proc.intf[-1], 'complete'):
            try:
                from readline import set_completer, parse_and_bind
                parse_and_bind("tab: complete")
                set_completer(self.proc.intf[-1].complete)
            except ImportError:
                pass
            pass

        if have_line_edit:
            self.proc.read_history_file()
            pass
        return
    pass

# if __name__ == '__main__':
#     from trepan import debugger as Mdebugger
#     d = Mdebugger.Debugger()
#     command = PythonCommand(d.core.processor)
#     command.proc.frame = sys._getframe()
#     command.proc.setup()
#     if len(sys.argv) > 1:
#         print("Type Python commands and exit to quit.")
#         print(sys.argv[1])
#         if sys.argv[1] == '-d':
#             print(command.run(['bpy', '-d']))
#         else:
#             print(command.run(['bpy']))
#             pass
#         pass
#     pass



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2017, 2020 Rocky Bernstein
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from trepan.processor.parse.semantics import build_location, build_range, Location
from trepan.processor.parse.parser import LocationError
from trepan.processor.parse.scanner import ScannerError
from trepan.processor.location import resolve_location

INVALID_PARSE_LIST = (None, None, None)
def parse_list_cmd(proc, args, listsize=10):
    """Parses arguments for the "list" command and returns the tuple:
    (filename, first line number, last line number)
    or sets these to None if there was some problem."""

    text = proc.current_command[len(args[0])+1:].strip()

    if text in frozenset(('', '.', '+', '-')):
        if text == '.':
            location = resolve_location(proc, '.')
            return location.path, location.line_number, listsize
        else:
            if proc.list_lineno is None:
                proc.errmsg("Don't have previous list location")
                return INVALID_PARSE_LIST
            filename = proc.list_filename
            if text == '+':
                first = max(1, proc.list_lineno + listsize)
            elif text == '-':
                if proc.list_lineno == 1 + listsize:
                    proc.errmsg("Already at start of %s." % proc.list_filename)
                    return INVALID_PARSE_LIST
                first = max(1, proc.list_lineno - (2*listsize) - 1)
            elif text == '':
                # Continue from where we last left off
                first = proc.list_lineno + 1
        last = first + listsize - 1
        return filename, first, last
    else:
        try:
            list_range = build_range(text)
        except LocationError as e:
            proc.errmsg("Error in parsing list range at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_LIST
        except ScannerError as e:
            proc.errmsg("Lexical error in parsing list range at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_LIST

        if list_range.first is None:
            # Last must have been given
            assert isinstance(list_range.last, Location)
            location = resolve_location(proc, list_range.last)
            if not location:
                return INVALID_PARSE_LIST
            last     = location.line_number
            first    = max(1, last - listsize)
            return location.path, first, last
        elif isinstance(list_range.first, int):
            first    = list_range.first
            location = resolve_location(proc, list_range.last)
            if not location:
                return INVALID_PARSE_LIST
            filename = location.path
            last     = location.line_number
            if last < first:
                # Treat as a count rather than an absolute location
                last = first + last
            return location.path, first, last
        else:
            # First is location. Last may be empty or a number
            assert isinstance(list_range.first, Location)
            location = resolve_location(proc, list_range.first)
            if not location:
                return INVALID_PARSE_LIST
            first    = location.line_number
            last     = list_range.last
            if location.method:
                first -= listsize // 2
            if isinstance(last, str):
                # Is an offset +number
                assert last[0] == '+'
                last = first + int(last[1:])
            elif not last:
                last = first + listsize
            elif last < first:
                # Treat as a count rather than an absolute location
                last = first + last

            return location.path, first, last
        pass
    return

INVALID_PARSE_LOCATION = (None, None)
def parse_location(proc, args):
    text = proc.current_command[len(args[0])+1:].strip()

    if text in frozenset(('', '.')):
        if text == '.':
            location = resolve_location(proc, '.')
            return location.path, location.line_number
        else:
            filename = proc.list_filename
            if text == '':
                # Continue from where we last left off
                first = proc.list_lineno + 1
        return filename, first
    else:
        try:
            location = build_location(text)
        except LocationError as e:
            proc.errmsg("Error in parsing location at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_LOCATION
        except ScannerError as e:
            proc.errmsg("Lexical error in location at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_LOCATION
        return location

# Demo it
if __name__=='__main__':
    from trepan.processor.command import mock as Mmock
    from trepan.processor.cmdproc import CommandProcessor
    import sys
    d = Mmock.MockDebugger()
    cmdproc = CommandProcessor(d.core)
    # print '-' * 10
    # print_source_line(sys.stdout.write, 100, 'source_line_test.py')
    # print '-' * 10
    cmdproc.frame = sys._getframe()
    cmdproc.setup()

    def five():
        return 5
    import os
    for cmd in (
            # "list",
            # "list +",
            # "list -",
            # "list 15, 10",
            "list five()",
            # "list 9 , 5",
            # "list 7 ,",
            # "list '''c:\\tmp\\foo.bat''':1",
            # 'list """/Users/My Documents/foo.py""":2',
            # 'list build_range()',
            # 'list os:1 ,',
            'list os.path:9 ,',
            ):
        args = cmd.split(' ')
        cmdproc.current_command = cmd
        print(parse_list_cmd(cmdproc, args))
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2009, 2013 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os, linecache
import os.path as osp


# FIXME: do a better job of this. Live parsing?
def is_ok_line_for_breakpoint(filename, lineno, errmsg_fn):
    """Check whether specified line seems to be executable.

    Return `lineno` if it is, 0 if not (e.g. a docstring, comment, blank
    line or EOF). Warning: testing is not comprehensive.
    """
    line = linecache.getline(filename, lineno)
    if not line:
        errmsg_fn('End of file')
        return False
    line = line.strip()
    # Don't allow setting breakpoint at a blank line
    if (not line or (line[0] == '#') or
         (line[:3] == '"""') or line[:3] == "'''"):
        errmsg_fn('Blank or comment')
        return False
    return True


def file2module(filename):
    """Given a file name, extract the most likely module name. """
    basename = osp.basename(filename)
    if '.' in basename:
        pos = basename.rfind('.')
        return basename[:pos]
    else:
        return basename
    return None


def search_file(filename, directories, cdir):
    """Return a full pathname for filename if we can find one. path
    is a list of directories to prepend to filename. If no file is
    found we'll return None"""

    for trydir in directories:

        # Handle $cwd and $cdir
        if trydir =='$cwd': trydir='.'
        elif trydir == '$cdir': trydir = cdir

        tryfile = osp.realpath(osp.join(trydir, filename))
        if osp.isfile(tryfile):
            return tryfile
    return None


def whence_file(py_script, dirnames=None):
    """Do a shell-like path lookup for py_script and return the results.
    If we can't find anything return py_script"""
    if py_script.find(os.sep) != -1:
        # Don't search since this name has path separator components
        return py_script
    if dirnames is None:
        dirnames = os.environ['PATH'].split(os.pathsep)
    for dirname in dirnames:
        py_script_try = osp.join(dirname, py_script)
        if osp.exists(py_script_try):
            return py_script_try
    # Failure
    return py_script

def path_expanduser_abs(filename):
    return os.path.abspath(os.path.expanduser(filename))

# Demo
if __name__=='__main__':
    import sys
    print(file2module(sys.argv[0]), sys.argv[0])
    ok = is_ok_line_for_breakpoint(__file__, 1, sys.stdout.write)
    print("\nCan stop at line 1? ", ok)
    ok = is_ok_line_for_breakpoint(__file__, 2, sys.stdout.write)
    print("\nCan stop at line 2? ", ok)
    print(path_expanduser_abs("./.trepan3krc"))
    print(path_expanduser_abs("~/.trepan3krc"))



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import inspect, os, sys

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor import cmdproc as Mcmdproc


class JumpCommand(DebuggerCommand):
    """**jump** *lineno*

Set the next line that will be executed. The line must be within the
stopped or bottom-most execution frame."""

    aliases       = ('j',)
    category      = 'running'
    execution_set = ['Running']
    min_args      = 1
    max_args      = 1
    name          = os.path.basename(__file__).split('.')[0]
    need_stack    = False
    short_help    = 'Set the next line to be executed'

    def run(self, args):
        if not self.core.is_running(): return False

        if self.proc.curindex + 1 != len(self.proc.stack):
            self.errmsg("You can only jump within the bottom frame")
            return False

        if self.proc.curframe.f_trace is None:
            self.errmsg("Sigh - operation can't be done here.")
            return False

        lineno = self.proc.get_an_int(args[1],
                                      ("jump: a line number is required, " +
                                       "got %s.") % args[1])
        if lineno is None: return False
        try:
            # Set to change position, update our copy of the stack,
            # and display the new position
            print(self.proc.curframe.f_trace)
            self.proc.curframe.f_lineno = lineno
            self.proc.stack[self.proc.curindex] = \
                self.proc.stack[self.proc.curindex][0], lineno
            Mcmdproc.print_location(self.proc)
        except ValueError as e:
            self.errmsg('jump failed: %s' % e)
        return False
    pass

if __name__ == '__main__':
    from trepan.processor.command import mock
    d, cp = mock.dbg_setup()
    command = JumpCommand(cp)
    print('jump when not running: ', command.run(['jump', '1']))
    command.core.execution_status = 'Running'
    cp.curframe = inspect.currentframe()
    cp.curindex = 0
    cp.stack = Mcmdproc.get_stack(cp.curframe, None, None)
    command.run(['jump', '1'])
    cp.curindex = len(cp.stack)-1
    command.run(['jump', '1'])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2009, 2013-2017, 2019-2021 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Some singleton debugger methods that can be called without first
creating a debugger object -- these methods will create a debugger object,
if necessary, first.
"""

# The following routines could be done via something like the following
# untested code:
# def _build_standalone(methodname, docstring=None):
#     def function(arg, arg2=None, globals=None, locals=None):
#         Debugger().getattr(methodname)(arg, arg2, globals, locals)
#        return
#    function.__name__ = methodname
#    function.__doc__ = docstring
#    return function
#
# But this a bit cumbersome and perhaps overkill for the 2 or so
# functions below.  (It also doesn't work once we add the exception handling
# we see below. So for now, we'll live with the code duplication.

import sys

from trepan import debugger as Mdebugger
from trepan.post_mortem import post_mortem_excepthook, uncaught_exception
from trepan.debugger import Trepan


def debugger_on_post_mortem():
    """Call debugger on an exeception that terminates a program"""
    sys.excepthook = post_mortem_excepthook
    return


def run_eval(
    expression,
    debug_opts=None,
    start_opts=None,
    globals_=None,
    locals_=None,
    tb_fn=None,
):

    """Evaluate the expression (given as a string) under debugger
    control starting with the statement subsequent to the place that
    this appears in your program.

    This is a wrapper to Debugger.run_eval(), so see that.

    When run_eval() returns, it returns the value of the expression.
    Otherwise this function is similar to run().
    """

    dbg = Trepan(opts=debug_opts)
    try:
        return dbg.run_eval(
            expression, start_opts=start_opts, globals_=globals_, locals_=locals_
        )
    except:
        dbg.core.trace_hook_suspend = True
        if start_opts and "tb_fn" in start_opts:
            tb_fn = start_opts["tb_fn"]
        uncaught_exception(dbg, tb_fn)
    finally:
        dbg.core.trace_hook_suspend = False
    return


def run_call(func, *args, **kwds):

    """Call the function (a function or method object, not a string)
    with the given arguments starting with the statement subsequent to
    the place that this appears in your program.

    When run_call() returns, it returns whatever the function call
    returned.  The debugger prompt appears as soon as the function is
    entered."""

    dbg = Trepan()
    try:
        return dbg.run_call(func, *args, **kwds)
    except:
        uncaught_exception(dbg)
        pass
    return


def run_exec(statement, debug_opts=None, start_opts=None, globals_=None, locals_=None):

    """Execute the statement (given as a string) under debugger
    control starting with the statement subsequent to the place that
    this run_call appears in your program.

    This is a wrapper to Debugger.run_exec(), so see that.

    The debugger prompt appears before any code is executed;
    you can set breakpoints and type 'continue', or you can step
    through the statement using 'step' or 'next'

    The optional globals_ and locals_ arguments specify the environment
    in which the code is executed; by default the dictionary of the
    module __main__ is used."""

    dbg = Trepan(opts=debug_opts)
    try:
        return dbg.run_exec(
            statement, start_opts=start_opts, globals_=globals_, locals_=locals_
        )
    except:
        uncaught_exception(dbg)
        pass
    return


def debug(dbg_opts=None, start_opts=None, post_mortem=True, step_ignore=1, level=0):
    """
Enter the debugger.

Parameters
----------

level : how many stack frames go back. Usually it will be
the default 0. But sometimes though there may be calls in setup to the debugger
that you may want to skip.

step_ignore : how many line events to ignore after the
debug() call. 0 means don't even wait for the debug() call to finish.

param dbg_opts : is an optional "options" dictionary that gets fed
trepan.Debugger(); `start_opts' are the optional "options"
dictionary that gets fed to trepan.Debugger.core.start().

Use like this:

.. code-block:: python

    ... # Possibly some Python code
    import trepan.api # Needed only once
    ... # Possibly some more Python code
    trepan.api.debug() # You can wrap inside conditional logic too
    pass  # Stop will be here.
    # Below is code you want to use the debugger to do things.
    ....  # more Python code
    # If you get to a place in the program where you aren't going
    # want to debug any more, but want to remove debugger trace overhead:
    trepan.api.stop()

Parameter "level" specifies how many stack frames go back. Usually it will be
the default 0. But sometimes though there may be calls in setup to the debugger
that you may want to skip.

Parameter "step_ignore" specifies how many line events to ignore after the
debug() call. 0 means don't even wait for the debug() call to finish.

In situations where you want an immediate stop in the "debug" call
rather than the statement following it ("pass" above), add parameter
step_ignore=0 to debug() like this::

    import trepan.api  # Needed only once
    # ... as before
    trepan.api.debug(step_ignore=0)
    # ... as before

Module variable _debugger_obj_ from module trepan.debugger is used as
the debugger instance variable; it can be subsequently used to change
settings or alter behavior. It should be of type Debugger (found in
module trepan). If not, it will get changed to that type::

   $ python
   >>> from trepan.debugger import debugger_obj
   >>> type(debugger_obj)
   <type 'NoneType'>
   >>> import trepan.api
   >>> trepan.api.debug()
   ...
   (Trepan) c
   >>> from trepan.debugger import debugger_obj
   >>> debugger_obj
   <trepan.debugger.Debugger instance at 0x7fbcacd514d0>
   >>>

If however you want your own separate debugger instance, you can
create it from the debugger _class Debugger()_ from module
trepan.debugger::

  $ python
  >>> from trepan.debugger import Debugger
  >>> dbgr = Debugger()  # Add options as desired
  >>> dbgr
  <trepan.debugger.Debugger instance at 0x2e25320>

`dbg_opts' is an optional "options" dictionary that gets fed
trepan.Debugger(); `start_opts' are the optional "options"
dictionary that gets fed to trepan.Debugger.core.start().
"""
    if not isinstance(Mdebugger.debugger_obj, Mdebugger.Trepan):
        Mdebugger.debugger_obj = Mdebugger.Trepan(dbg_opts)
        Mdebugger.debugger_obj.core.add_ignore(debug, stop)
        pass
    core = Mdebugger.debugger_obj.core
    frame = sys._getframe(0 + level)
    core.set_next(frame)
    if start_opts and "startup-profile" in start_opts and start_opts["startup-profile"]:
        dbg_initfiles = start_opts["startup-profile"]
        from trepan import options

        options.add_startup_file(dbg_initfiles)
        for init_cmdfile in dbg_initfiles:
            core.processor.queue_startfile(init_cmdfile)

    if not core.is_started():
        core.start(start_opts)
        pass
    if post_mortem:
        debugger_on_post_mortem()
        pass
    if 0 == step_ignore:
        frame = sys._getframe(1 + level)
        core.stop_reason = "at a debug() call"
        old_trace_hook_suspend = core.trace_hook_suspend
        core.trace_hook_suspend = True
        core.processor.event_processor(frame, "line", None)
        core.trace_hook_suspend = old_trace_hook_suspend
    else:
        core.step_ignore = step_ignore - 1
        pass
    return


def stop(opts=None):
    if isinstance(Mdebugger.debugger_obj, Trepan):
        return Mdebugger.debugger_obj.stop(opts)
    return None


# # Demo it
if __name__ == "__main__":
    import tracer

    def foo(n):
        y = n
        for i in range(n):
            print(i)
            pass
        return y

    from trepan.lib.default import DEBUGGER_SETTINGS

    settings = dict(DEBUGGER_SETTINGS)
    settings.update({"trace": True, "printset": tracer.ALL_EVENTS})
    debug_opts = {"step_ignore": -1, "settings": settings}
    print('Issuing: run_eval("1+2")')
    print(run_eval("1+2", debug_opts=debug_opts))
    print('Issuing: run_exec("x=1; y=2")')
    run_exec("x=1; y=2", debug_opts=debug_opts)
    print("Issuing: run_call(foo, debug_opts, None, 2)")
    if len(sys.argv) > 1:
        print(
            "Issuing interactive: run_call(foo, %s, None, %s)"
            % (debug_opts, sys.argv[1])
        )
        run_call(foo, debug_opts, int(sys.argv[1]))
    else:
        run_call(foo, debug_opts, 2)
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009-2017, 2021 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Subsidiary routines used to "pack" and "unpack" TCP messages. """

TCP_MAX_PACKET = 8192  # Largest size for a recv
LOG_MAX_MSG = 4  # int(log(TCP_MAX_PACKET)


def pack_msg(msg):
    fmt = "%%0%dd" % LOG_MAX_MSG  # A funny way of writing: '%04d'
    byte_msg = bytes(msg, "UTF-8")
    byte_fmt = bytes(fmt % len(byte_msg), "UTF-8")
    return byte_fmt + byte_msg


def unpack_msg(buf):
    if len(buf) == 0:
        # Fake a quit
        return "", bytes("q".encode("utf-8"))
    length = int(buf[0:LOG_MAX_MSG])
    data = buf[LOG_MAX_MSG : LOG_MAX_MSG + length]
    buf = buf[LOG_MAX_MSG + length :]
    return buf, data


# Demo
if __name__ == "__main__":
    print(unpack_msg(pack_msg("Hello, there!"))[1])
    # assert unpack_msg(pack_msg(msg))[1] == msg
    pass



#!/usr/bin/env python3
'Unit test for trepan.processor.cmdproc'
import os, sys, unittest

from trepan.lib import core as Mcore


class MockProcessor:
    pass


class TestCore(unittest.TestCase):

    def test_is_next_stop(self):
        # We assume there's at least one command
        opts = {'processor': MockProcessor()}
        dc = Mcore.TrepanCore(None, opts=opts)
        dc.step_ignore = 1
        self.assertFalse(dc._is_step_next_stop('line'))
        self.assertEqual(0, dc.step_ignore)
        self.assertTrue(dc._is_step_next_stop('line'))
        self.assertEqual(0, dc.step_ignore)

        return

    def test_canonic(self):
        opts = {'processor': MockProcessor()}
        dc = Mcore.TrepanCore(None, opts=opts)
        s = '<string>'
        self.assertEqual(s, dc.canonic(s),
                         'canonic should not have changed string')
        if sys.platform != 'win32':
            self.assertEqual(os.path.sep, dc.canonic(__file__)[0],
                             'canonic should produce an absolute file')
        return

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#   Copyright (C) 2015-2016, 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pyficache

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class SetSubstitute(Mbase_subcmd.DebuggerSubcommand):
    """**set substitute** *from-name* *to-path*

Add a substitution rule replacing *from-name* into *to-path* in source file names.
If a substitution rule was previously set for *from-name*, the old rule
is replaced by the new one.

Spaces in "filenames" like `<frozen importlib._bootstrap>` messes up our normal shell
tokenization, so we have added a hack to ignore `<frozen .. >`.

So, for frozen files like `<frozen importlib._bootstrap>`, `use importlib._bootstrap`

Examples:
---------

    set substitute importlib._bootstrap /usr/lib/python3.4/importlib/_bootstrap.py
    set substitute ./gcd.py /tmp/gcd.py

"""

    in_list = True
    min_abbrev = len("sub")
    short_help = "Set filename substitution"

    def run(self, args):
        pyficache.remap_file(args[1], args[0])

    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    Mhelper.demo_run(SetSubstitute)
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009-2010, 2012-2013, 2015-2016, 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""A base class for debugger subcommands.

This file is a module in this directory that isn't a real command
and commands.py needs to take care to avoid instantiating this class
and storing it as a list of known debugger commands.
"""

NotImplementedMessage = "This method must be overriden in a subclass"
import columnize, re
from pygments.console import colorize


# Note: don't end classname with Command (capital C) since cmdproc
# will think this a command name like QuitCommand
#                                         ^
class DebuggerSubcommand:
    """Base Class for Debugger subcommands. We pull in some helper
    functions for command from module cmdfns."""

    in_list = True  # Show item in help list of commands

    # Run subcommand for those subcommands like "show"
    # which append current settings to list output.
    run_cmd = True

    run_in_help = True  # Run to get value in 'show' command?
    min_abbrev = 1
    min_args = 0
    max_args = None
    need_stack = False

    def __init__(self, cmd):
        """cmd contains the command object that this
        command is invoked through.  A debugger field gives access to
        the stack frame and I/O."""
        self.cmd = cmd

        # Convenience class access. We don't expect that any of these
        # will change over the course of the program execution like
        # errmsg(), msg(), and msg_nocr() might. (See the note below
        # on these latter 3 methods.)
        #
        self.proc = cmd.proc
        self.core = cmd.core
        self.debugger = cmd.debugger
        self.settings = cmd.debugger.settings

        if not hasattr(self, "short_help"):
            help = self.__doc__.split("\n")
            if len(help) > 0:
                if help[0][0] == "*" and len(help) > 2:
                    self.short_help = help[2]
                else:
                    self.short_help = help[0]
            pass

        # By default the name of the subcommand will be the name of the
        # last part of module (e.g. "args" in "infos.args" or "basename"
        # in "shows.basename"). However it *is* possible for one to change
        # that -- perhaps one may want to put several subcommands into
        # a single file. So in those cases, one will have to set self.name
        # accordingly by other means.
        self.name = self.__module__.split(".")[-1]

        return

    def columnize_commands(self, commands):
        """List commands arranged in an aligned columns"""
        commands.sort()
        width = self.debugger.settings["width"]
        return columnize.columnize(commands, displaywidth=width, lineprefix="    ")

    def confirm(self, msg, default=False):
        """ Convenience short-hand for self.debugger.intf.confirm """
        return self.debugger.intf[-1].confirm(msg, default)

    # Note for errmsg, msg, and msg_nocr we don't want to simply make
    # an assignment of method names like self.msg = self.debugger.intf.msg,
    # because we want to allow the interface (intf) to change
    # dynamically. That is, the value of self.debugger may change
    # in the course of the program and if we made such an method assignemnt
    # we wouldn't pick up that change in our self.msg
    def errmsg(self, msg):
        """ Convenience short-hand for self.debugger.intf[-1].errmsg """
        return self.debugger.intf[-1].errmsg(msg)

    def msg(self, msg):
        """ Convenience short-hand for self.debugger.intf[-1].msg """
        return self.debugger.intf[-1].msg(msg)

    def msg_nocr(self, msg):
        """ Convenience short-hand for self.debugger.intf[-1].msg_nocr """
        return self.debugger.intf[-1].msg_nocr(msg)

    aliases = ()
    name = "YourCommandName"

    def rst_msg(self, text):
        """Convenience short-hand for self.proc.rst_msg(text)"""
        return self.proc.rst_msg(text)

    def run(self):
        """ The method that implements the debugger command.
        Help on the command comes from the docstring of this method.
        """
        raise NotImplementedError(NotImplementedMessage)

    def section(self, message, opts={}):
        if "plain" != self.settings["highlight"]:
            message = colorize("bold", message)
        else:
            message += "\n" + "-" * len(message)
            pass
        self.msg(message)

    pass


from trepan.processor.cmdfns import run_set_bool, run_show_bool, run_show_int
from trepan.lib.complete import complete_token


class DebuggerSetBoolSubcommand(DebuggerSubcommand):
    max_args = 1

    def complete(self, prefix):
        result = complete_token(("on", "off"), prefix)
        return result

    def run(self, args):
        # Strip off ReStructuredText tags
        doc = re.sub("[*]", "", self.short_help).lstrip()
        # Take only the first two tokens
        i = doc.find(" ")
        if i > 0:
            j = doc.find(" ", i + 1)
            if j > 0:
                doc = doc[0:j]
            pass
        doc = doc.capitalize().split("\n")[0].rstrip(".")
        run_set_bool(self, args)
        run_show_bool(self, doc)
        return

    def summary_help(self, subcmd_name, subcmd):
        return self.msg_nocr("%-12s: " % self.short_help)

    pass


class DebuggerShowIntSubcommand(DebuggerSubcommand):
    max_args = 0

    def run(self, args):
        if hasattr(self, "short_help"):
            short_help = self.short_help
        else:
            short_help = self.__doc__[5:].capitalize()
            pass
        run_show_int(self, short_help)
        return


class DebuggerShowBoolSubcommand(DebuggerSubcommand):
    max_args = 0

    def run(self, args):
        # Strip off ReStructuredText tags
        doc = re.sub("[*]", "", self.short_help)
        doc = doc[5:].capitalize().split("\n")[0].rstrip(".")
        run_show_bool(self, doc)
        return


if __name__ == "__main__":
    from trepan.processor.command.mock import dbg_setup

    d, cp = dbg_setup()
    dd = DebuggerSubcommand(cp.commands["quit"])



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013-2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os, sys

from trepan.processor.command.base_cmd import DebuggerCommand


class CDCommand(DebuggerCommand):
    """**cd** *directory*

Set working directory to *directory* for debugger and program
being debugged. """

    aliases = ("chdir",)
    short_help = (
        "Set working directory to DIR for debugger " "and program being debugged"
    )

    DebuggerCommand.setup(
        locals(), category="files", max_args=1, min_args=1,
    )

    def run(self, args):
        try:
            os.chdir(args[1])
            self.msg("Working directory %s." % os.getcwd())
        except OSError:
            self.errmsg("cd: %s." % sys.exc_info()[1])
            pass
        return

    pass


if __name__ == "__main__":
    from trepan import debugger as Mdebugger

    d = Mdebugger.Trepan()
    cmd = CDCommand(d.core.processor)
    for c in (
        ["cd", "wrong", "number", "of", "args"],
        ["cd", "foo"],
        ["cd", "."],
        ["cd", "/"],
    ):
        cmd.run(c)
        pass
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013-2017, 2021 Rocky Bernstein
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#    02110-1301 USA.

import sys, time

# Our local modules
from trepan.interfaces import client as Mclient
from trepan.interfaces import comcodes as Mcomcodes

from optparse import OptionParser
from trepan.version import __version__


def process_options(pkg_version, sys_argv, option_list=None):
    """Handle debugger options. Set `option_list' if you are writing
    another main program and want to extend the existing set of debugger
    options.

    The options dicionary from opt_parser is return. sys_argv is
    also updated."""
    usage_str = """%prog [debugger-options]]

    Client connection to an out-of-process trepan3k debugger session"""

    # serverChoices = ('TCP','FIFO', None) # we use PID for now.

    optparser = OptionParser(
        usage=usage_str,
        option_list=option_list,
        version="%%prog version %s" % pkg_version,
    )

    optparser.add_option(
        "-H",
        "--host",
        dest="host",
        default="127.0.0.1",
        action="store",
        type="string",
        metavar="IP-OR-HOST",
        help="connect IP or host name.",
    )
    optparser.add_option(
        "-P",
        "--port",
        dest="port",
        default=1027,
        action="store",
        type="int",
        metavar="NUMBER",
        help="Use TCP port number NUMBER for " "out-of-process connections.",
    )
    optparser.add_option(
        "--pid",
        dest="pid",
        default=0,
        action="store",
        type="int",
        metavar="NUMBER",
        help="Use PID to get FIFO names for " "out-of-process connections.",
    )

    optparser.disable_interspersed_args()

    sys.argv = list(sys_argv)
    (opts, sys.argv) = optparser.parse_args()
    return opts, sys.argv


#
# Connects to a debugger in server mode
#

# DEFAULT_CLIENT_CONNECTION_OPTS = {'open': True, 'IO': 'FIFO'}
DEFAULT_CLIENT_CONNECTION_OPTS = {
    "open": True,
    "IO": "TCP",
    "HOST": "127.0.0.1",
    "PORT": 1027,
}


def start_client(connection_opts):
    intf = Mclient.ClientInterface(connection_opts=connection_opts)
    # debugger.interface.append(intf)
    intf.msg("Connected.")
    done = False
    while not done:
        control, remote_msg = intf.read_remote()
        # print 'c, r', control, remote_msg
        if Mcomcodes.PRINT == control:
            print(remote_msg, end=" ")
            pass
        elif control in [Mcomcodes.CONFIRM_TRUE, Mcomcodes.CONFIRM_FALSE]:
            default = Mcomcodes.CONFIRM_TRUE == control
            if intf.confirm(remote_msg.rstrip("\n"), default):
                msg = "Y"
            else:
                msg = "N"
                pass
            intf.write_remote(Mcomcodes.CONFIRM_REPLY, msg)
        elif Mcomcodes.PROMPT == control:
            msg = intf.read_command("(Trepan*) ").strip()
            intf.write_remote(Mcomcodes.CONFIRM_REPLY, msg)
        elif Mcomcodes.QUIT == control:
            print("trepan3kc: That's all, folks...")
            done = True
            break
        elif Mcomcodes.RESTART == control:
            # FIXME need to save stuff like port # and
            # and for FIFO we need new pid.
            if "TCP" == connection_opts["IO"]:
                print("Restarting...")
                intf.inout.close()
                time.sleep(1)
                intf.inout.open()
            else:
                print("Don't know how to hard-restart FIFO...")
                done = True
                pass
            break
        else:
            print("!! Weird status code received '%s'" % control)
            print(remote_msg, end=" ")
            pass
        pass
    intf.close()
    return


def run(opts, sys_argv):
    if hasattr(opts, "pid") and opts.pid > 0:
        remote_opts = {"open": opts.pid, "IO": "FIFO"}
    else:
        remote_opts = {"open": True, "IO": "TCP", "PORT": opts.port, "HOST": opts.host}
    start_client(remote_opts)
    return


def main():
    opts, sys_argv = process_options(__version__, sys.argv)
    # print(opts)
    run(opts, sys_argv)
    return


if __name__ == "__main__":
    main()
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009-2010, 2013, 2015, 2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand
from trepan.processor import cmdfns as Mcmdfns


class SetMaxString(DebuggerSubcommand):
    """**set maxstring** *number*

Set the number of characters allowed in showing string values

See also:
---------

`show maxstring`
"""

    in_list = True
    max_args = 1
    min_abbrev = len("str")  # Need at least "set max str"
    min_args = 1
    short_help = "Set maximum characters in showing strings"

    def run(self, args):
        Mcmdfns.run_set_int(
            self,
            " ".join(args),
            "The '%s' command requires a character count" % self.name,
            0,
            None,
        )
        self.proc._repr.maxstring = self.settings[self.name]
        return None

    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2009, 2012-2013, 2015 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
import columnize, inspect, pyficache, sys

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand
from trepan import misc as Mmisc
from trepan.lib import complete as Mcomplete
from trepan.lib.file import file_list


class InfoFiles(DebuggerSubcommand):
    min_abbrev = 2
    need_stack = False
    short_help = 'Show information about an imported or loaded Python file'

    def complete(self, prefix):
        completions = sorted(['.'] + file_list())
        return Mcomplete.complete_token(completions, prefix)

    def run(self, args):
        """**info files** [*filename* [**all** | **brkpts** | **lines** | **sha1** | **size**]]

Show information about the current file. If no filename is
given and the program is running then the current file associated
with the current stack entry is used. Sub options which can be
shown about a file are:

* **brkpts** Line numbers where there are statement boundaries. These lines can be used in breakpoint commands.

* **sha1**	A SHA1 hash of the source text.

The following may be useful in comparing source code.

* **size**	The number of lines in the file.

* **all** All of the above information.
        """
        if len(args) == 0:
            if not self.proc.curframe:
                self.errmsg("No frame - no default file.")
                return False
            filename = self.proc.curframe.f_code.co_filename
        else:
            filename = args[0]
            pass

        m = filename + ' is'
        filename_cache = self.core.filename_cache
        if filename in filename_cache:
            m += " cached in debugger"
            if filename_cache[filename] != filename:
                m += ' as:'
                m = Mmisc.wrapped_lines(m, filename_cache[filename] + '.',
                                        self.settings['width'])
            else:
                m += '.'
                pass
            self.msg(m)
        else:
            matches = [file for file in file_list() if
                       file.endswith(filename)]
            if (len(matches) > 1):
                self.msg("Multiple files found ending filename string:")
                for match_file in matches:
                    self.msg("\t%s" % match_file)
                    pass
            elif len(matches) == 1:
                canonic_name = pyficache.unmap_file(matches[0])
                m += " matched debugger cache file:\n  "  + canonic_name
                self.msg(m)
            else:
                self.msg(m + ' not cached in debugger.')
            pass
        canonic_name = self.core.canonic(filename)
        self.msg(Mmisc.wrapped_lines('Canonic name:', canonic_name,
                                     self.settings['width']))
        for name in (canonic_name, filename):
            if name in sys.modules:
                for key in [k for k, v in list(sys.modules.items())
                            if name == v]:
                    self.msg("module: %s", key)
                    pass
                pass
            pass
        for arg in args[1:]:
            processed_arg = False
            if arg in ['all', 'size']:
                if pyficache.size(canonic_name):
                    self.msg("File has %d lines." %
                             pyficache.size(canonic_name))
                    pass
                processed_arg = True
                pass
            if arg in ['all', 'sha1']:
                self.msg("SHA1 is %s." % pyficache.sha1(canonic_name))
                processed_arg = True
                pass
            if arg in ['all', 'brkpts']:
                lines = pyficache.trace_line_numbers(canonic_name)
                if lines:
                    self.section("Possible breakpoint line numbers:")
                    fmt_lines = columnize.columnize(list(lines), ljust = False,
                                                    arrange_vertical = False,
                                                    lineprefix='  ')
                    self.msg(fmt_lines)
                    pass
                processed_arg = True
                pass
            if not processed_arg:
                self.errmsg("Don't understand sub-option %s." % arg)
                pass
            pass
        return
    pass

if __name__ == '__main__':
    from trepan.processor.command import mock, info as Minfo
    d, cp = mock.dbg_setup()
    i = Minfo.InfoCommand(cp)
    sub = InfoFiles(i)
    sub.run([])
    cp.curframe = inspect.currentframe()
    sub.run(['file.py', 'foo'])
    for width in (200, 80):
        sub.settings['width'] = width
        sub.run(['file.py', 'lines'])
        print(sub.run([]))
        pass
    sub.run(['file.py', 'all'])
    print(sub.complete(''))
    # sub.run(['file.py', 'lines', 'sha1'])
    pass



"""
Simple SPARK-style scanner
Copyright (c) 2017-2018 Rocky Bernstein
"""

from __future__ import print_function
import re
from spark_parser.scanner import GenericScanner
from trepan.processor.parse.tok import Token

class ScannerError(Exception):
    def __init__(self, text, text_cursor):
        self.text = text
        self.text_cursor = text_cursor

    def __str__(self):
        return self.text + "\n" + self.text_cursor

class LocationScanner(GenericScanner):

    def error(self, s):
        """Show text and a caret under that. For example:
x = 2y + z
     ^
"""
        # print("Lexical error:")
        # print("%s" % s[:self.pos+10])  # + 10 for trailing context
        # print("%s^" % (" "*(self.pos-1)))
        # for t in self.rv: print(t)
        raise ScannerError( ("%s" % s),
                            ("%s^" % (" "*(self.pos-1))) )

    def tokenize(self, input):
        self.rv = []
        GenericScanner.tokenize(self, input)
        return self.rv

    def add_token(self, name, v):
        t = Token(kind=name, value=v, offset=self.pos)
        self.pos += len(str(v))
        self.rv.append(t)

    # The function names below begin with 't_'.
    # This indicates to GenericScanner that these routines
    # form the tokens. GenericScanner introspects on the
    # method names of this class and the docstrings to come
    # up with both the names of the tokens and the regular expressions
    # that make up those tokens

    # Recognize white space, but we don't create a token for it.
    # This has the effect of stripping white space between tokens
    def t_whitespace(self, s):
        r'\s+'
        self.add_token('SPACE', s)
        self.pos += len(s)
        pass

    def t_file_or_func(self, s):
        r'(?:[^*-+@,\d\'"\t \n:][^\'"\t \n:,]*)|(?:^""".+""")|(?:\'\'\'.+\'\'\')'
        maybe_funcname = True
        if s == 'if':
            self.add_token('IF', s)
            return
        if s[0] in frozenset(('"', "'")):
            # Pick out text inside of triple-quoted string
            if ( (s.startswith("'''") and s.endswith("'''") ) or
                 (s.startswith('"""') and s.endswith('"""') ) ):
                base = s[3:-3]
            else:
                # Pick out text inside singly-quote string
                base = s[1:-1]
            maybe_funcname = False
        else:
            base = s
        if maybe_funcname and re.match('[a-zA-Z_][a-zA-Z_.0-9]+\(\)', s):
            self.add_token('FUNCNAME', base)
        else:
            self.add_token('FILENAME', base)
        self.pos += len(s)

    def t_single_quote_file(self, s):
        r"'[^'].+'"
        # Pick out text inside of singe-quoted string
        base = s[1:-1]
        self.add_token('FILENAME', base)
        self.pos += len(s)

    def t_double_quote_file(self, s):
        r'"[^"]+"'
        # Pick out text inside of singe-quoted string
        base = s[1:-1]
        self.add_token('FILENAME', base)
        self.pos += len(s)

    def t_colon(self, s):
        r':'
        # Used to separate a filename from a line number
        self.add_token('COLON', s)
        self.pos += len(s)

    def t_comma(self, s):
        r','
        # Used in "list" to separate first from last
        self.add_token('COMMA', s)
        self.pos += len(s)

    def t_direction(self, s):
        r'^[+-]$'
        # Used in the "list" command
        self.add_token('DIRECTION', s)
        self.pos += len(s)

    # Recognize integers
    def t_number(self, s):
        r'\d+'
        pos = self.pos
        self.add_token('NUMBER', int(s))
        self.pos = pos + len(s)

    # Recognize list offsets (counts)
    def t_offset(self, s):
        r'[+]\d+'
        pos = self.pos
        self.add_token('OFFSET', s)
        self.pos = pos + len(s)

    # Recognize addresses (bytecode offsets)
    def t_address(self, s):
        r'[*@]\d+'
        pos = self.pos
        self.add_token('ADDRESS', s)
        self.pos = pos + len(s)

if __name__ == "__main__":
    for line in (
            # '/tmp/foo.py:12',
            # "'''/tmp/foo.py:12'''",
            "'/tmp/foo.py:12'",
            "6",
            "*6",
            "@6",
            "8 *6",
            # "/tmp/foo.py line 12",
            # "\"\"\"/tmp/foo.py's line 12\"\"\"",
            # "12",
            # "../foo.py:5",
            # "gcd()",
            # "foo.py line 5 if x > 1",
            # "5 ,",
            # "5,",
            # "5,10",
            # ",10",
            ):
        try:
            tokens = LocationScanner().tokenize(line.strip())
            for t in tokens:
                print(t)
                pass
        except ScannerError as e:
            print("Lexical error at or around: ")
            print(e.text)
            print(e.text_cursor)

        pass
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2020, 2021 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand
from trepan.lib.complete import complete_token

choices = ["classic", "extended", "extended-bytes", "bytes"]


class SetAsmFmt(DebuggerSubcommand):
    """**set asmfmt** {**classic** | **extended** | **bytes** | **extended-bytes**}

    The style of format to use in disassembly:

           classic:  fields: line, marker offset, opcode operand
           extended: above, but we try harder to get operand information from previous instructions
           bytes:  like classic but we show the instruction bytes after the offset
           extended-bytes:   bytes + extended


    Examples:
    --------

        set asmfmt extended # this is the default
        set asmfmt classic  # no highlight

    See also:
    ---------
    `show asmfmt`"""

    in_list = True
    max_args = 1
    min_abbrev = len("asmf")
    min_args = 0
    short_help = "Set disassembly style"

    def complete(self, prefix):
        return complete_token(choices, prefix)

    def get_format_type(self, arg):
        if not arg:
            return "extended"
        if arg in choices:
            return arg
        else:
            self.errmsg("Expecting one of: %s; got: %s." % (", ".join(choices), arg))
            return None
        pass

    def run(self, args):
        if len(args) == 0:
            self.section("disasembly style types: ")
            self.msg(self.columnize_commands(choices))
            return
        format_type = self.get_format_type(args[0])
        if not format_type:
            return
        self.debugger.settings[self.name] = format_type
        show_cmd = self.proc.commands["show"]
        show_cmd.run(["show", self.name])
        return

    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd.__demo_helper__ import demo_run

    demo_run(SetAsmFmt, ["classic"])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2009, 2015, 2018, 2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
""" Copyright (C) 2008-2009, 2015, 2018, 2020 Rocky Bernstein <rocky@gnu.org> """

import glob
import os.path as osp

# FIXME: Is it really helpful to "privatize" variable names below?
# The below names are not part of the standard pre-defined names like
# __name__ or __file__ are.

# Get the name of our directory.
__command_dir__ = osp.dirname(__file__)

# A glob pattern that will get all *.py files but not __init__.py
__py_files__ = glob.glob(osp.join(__command_dir__, "[a-z]*.py"))

# Take the basename of the filename and drop off '.py'. That becomes
# the list of modules that commands.py will use to import
exclude_files = []
__modules__ = [
    osp.basename(filename[0:-3])
    for filename in __py_files__
    if osp.basename(filename) not in exclude_files
]



from xdis import IS_PYPY
import difflib, os, re, sys, time
import os.path as osp
srcdir = osp.abspath(osp.dirname(__file__))


def run_debugger(testname, python_file, dbgr_opts='', args='',
                 outfile=None, right_template=None):
    datadir   = osp.join(srcdir, '..', 'data')
    progdir   = osp.join(srcdir, '..', 'example')
    dbgrdir   = osp.join(srcdir, '..', '..', 'trepan')
    dbgr_short= "__main__.py"
    dbgr_path = osp.join(dbgrdir, dbgr_short)

    if not right_template:
        if IS_PYPY:
            right_template = "%s-pypy.right"
        else:
            right_template = "%s.right"

    rightfile = osp.join(datadir, right_template % testname)

    sys.path.insert(0, osp.join(srcdir, '..', '..'))
    os.environ['PYTHONPATH'] = os.pathsep.join(sys.path)
    cmdfile     = osp.join(datadir, "%s.cmd"   % testname)
    outfile     = osp.join(srcdir, "%s.out" % testname)
    if python_file:
        programfile = osp.join(progdir, python_file)
    else:
        programfile = ''
        pass

    outfile_opt = '--output=%s ' % outfile

    if osp.exists(outfile): os.unlink(outfile)

    cmd = "%s --command %s %s %s %s %s" % \
          (dbgr_path, cmdfile, outfile_opt, dbgr_opts, programfile, args)

    # print(cmd)
    os.system(cmd)
    fromfile  = rightfile
    fromdate  = time.ctime(os.stat(fromfile).st_mtime)
    tofile    = outfile
    todate    = time.ctime(os.stat(tofile).st_mtime)
    with open(fromfile) as f: fromlines = f.readlines()
    with open(tofile) as f: tolines = f.readlines()

    # Filter out last instruction. For example:
    # (gcd.py:11 @6): -> (gcd.py:11)
    tolines = [re.sub(' @\d+\):', '):', line) for line in tolines]

    diff = list(difflib.unified_diff(fromlines, tolines, fromfile,
                                     tofile, fromdate, todate))
    if len(diff) == 0:
        os.unlink(outfile)
        pass
    else:
        with open(tofile, 'w') as out:
            out.writelines(tolines)
            pass
        pass
    for line in diff:
        print(line.rstrip())
        pass
    return len(diff) == 0
